from __future__ import annotations

import ast
from dataclasses import dataclass
from pathlib import Path

from .common import REPO_ROOT, SRC_ROOT, repo_relative


@dataclass(frozen=True)
class ModuleSummary:
    imports: tuple[str, ...]
    direct_importers: tuple[str, ...]
    transitive_importers: tuple[str, ...]
    impacted_packages: tuple[str, ...]


@dataclass(frozen=True)
class ImportGraph:
    imports: dict[str, tuple[str, ...]]
    reverse_imports: dict[str, tuple[str, ...]]
    module_paths: dict[str, str]

    def transitive_importers(self, module: str) -> tuple[str, ...]:
        visited: set[str] = set()
        stack = list(self.reverse_imports.get(module, ()))
        while stack:
            current = stack.pop()
            if current in visited:
                continue
            visited.add(current)
            stack.extend(self.reverse_imports.get(current, ()))
        return tuple(sorted(visited))

    def module_summary(self, module: str) -> ModuleSummary:
        direct_importers = self.reverse_imports.get(module, ())
        transitive_importers = self.transitive_importers(module)
        impacted_packages = sorted(
            {
                module_to_package(importer)
                for importer in transitive_importers
                if module_to_package(importer)
                not in {module_to_package(module), "dagzoo", "dagzoo.__main__"}
            }
        )
        return ModuleSummary(
            imports=self.imports.get(module, ()),
            direct_importers=direct_importers,
            transitive_importers=transitive_importers,
            impacted_packages=tuple(impacted_packages),
        )


def module_to_package(module: str) -> str:
    parts = module.split(".")
    if len(parts) <= 2:
        return module
    return ".".join(parts[:2])


def module_name_for_path(path: Path) -> str:
    relative = path.relative_to(SRC_ROOT)
    module = "dagzoo." + ".".join(relative.parts)
    module = module[:-3] if module.endswith(".py") else module
    if module.endswith(".__init__"):
        module = module[:-9]
    return module


def path_to_module(path_str: str) -> str | None:
    path = REPO_ROOT / path_str
    if not path.is_relative_to(SRC_ROOT) or path.suffix != ".py":
        return None
    return module_name_for_path(path)


def build_import_graph() -> ImportGraph:
    module_paths: dict[str, str] = {}
    known_modules: set[str] = set()
    for path in sorted(SRC_ROOT.rglob("*.py")):
        if "__pycache__" in path.parts:
            continue
        module = module_name_for_path(path)
        module_paths[module] = repo_relative(path)
        known_modules.add(module)

    imports: dict[str, tuple[str, ...]] = {}
    reverse_imports: dict[str, set[str]] = {module: set() for module in known_modules}
    for module, path_str in module_paths.items():
        path = REPO_ROOT / path_str
        module_imports = tuple(sorted(_collect_imports(path, module, known_modules)))
        imports[module] = module_imports
        for dependency in module_imports:
            reverse_imports.setdefault(dependency, set()).add(module)

    return ImportGraph(
        imports=imports,
        reverse_imports={key: tuple(sorted(value)) for key, value in reverse_imports.items()},
        module_paths=module_paths,
    )


def _collect_imports(path: Path, module: str, known_modules: set[str]) -> set[str]:
    tree = ast.parse(path.read_text())
    current_package = module if path.name == "__init__.py" else module.rsplit(".", 1)[0]
    discovered: set[str] = set()
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                normalized = _normalize_import(alias.name, known_modules)
                if normalized is not None and normalized != module:
                    discovered.add(normalized)
        elif isinstance(node, ast.ImportFrom):
            for dependency in _normalize_import_from(node, current_package, known_modules):
                if dependency != module:
                    discovered.add(dependency)
    return discovered


def _normalize_import(candidate: str, known_modules: set[str]) -> str | None:
    matches = [
        module
        for module in known_modules
        if candidate == module or candidate.startswith(module + ".")
    ]
    if not matches:
        return None
    return max(matches, key=len)


def _normalize_import_from(
    node: ast.ImportFrom, current_package: str, known_modules: set[str]
) -> set[str]:
    if node.level == 0:
        base_parts = ()
    else:
        current_parts = current_package.split(".")
        if node.level == 1:
            base_parts = tuple(current_parts)
        else:
            base_parts = tuple(current_parts[: -(node.level - 1)])
    module_parts = tuple(node.module.split(".")) if node.module else ()
    base_module = ".".join(base_parts + module_parts)
    discovered: set[str] = set()
    if base_module.startswith("dagzoo"):
        normalized = _normalize_import(base_module, known_modules)
        if normalized is not None:
            discovered.add(normalized)
    for alias in node.names:
        if alias.name == "*":
            continue
        child_candidate = f"{base_module}.{alias.name}" if base_module else alias.name
        normalized_child = _normalize_import(child_candidate, known_modules)
        if normalized_child is not None:
            discovered.add(normalized_child)
    return discovered



from __future__ import annotations

from copy import deepcopy
from typing import Any

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from dagzoo.io.lineage_schema import (
    LINEAGE_ADJACENCY_ENCODING,
    LINEAGE_SCHEMA_NAME,
    LINEAGE_SCHEMA_VERSION,
    LINEAGE_SCHEMA_VERSION_COMPACT,
    LineageValidationError,
    validate_lineage_payload,
    validate_metadata_lineage,
)

_N_NODES_STRATEGY = st.integers(min_value=2, max_value=12)
_FEATURE_COUNT_STRATEGY = st.integers(min_value=0, max_value=32)
_PATH_STRATEGY = st.text(
    alphabet=tuple("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789._/-"),
    min_size=1,
    max_size=64,
)
_HEX_SHA256_STRATEGY = st.text(
    alphabet=tuple("0123456789abcdef"),
    min_size=64,
    max_size=64,
)


def _upper_triangle_capacity(n_nodes: int) -> int:
    return (n_nodes * (n_nodes - 1)) // 2


def _dense_adjacency_from_bits(n_nodes: int, upper_bits: list[int]) -> list[list[int]]:
    adjacency = [[0 for _ in range(n_nodes)] for _ in range(n_nodes)]
    idx = 0
    for row in range(n_nodes):
        for col in range(row + 1, n_nodes):
            adjacency[row][col] = int(upper_bits[idx])
            idx += 1
    return adjacency


@st.composite
def _dense_lineage_payload_strategy(draw: st.DrawFn) -> dict[str, Any]:
    n_nodes = draw(_N_NODES_STRATEGY)
    adjacency = _dense_adjacency_from_bits(
        n_nodes,
        draw(
            st.lists(
                st.integers(min_value=0, max_value=1),
                min_size=_upper_triangle_capacity(n_nodes),
                max_size=_upper_triangle_capacity(n_nodes),
            )
        ),
    )
    feature_count = draw(_FEATURE_COUNT_STRATEGY)
    feature_to_node = draw(
        st.lists(
            st.integers(min_value=0, max_value=n_nodes - 1),
            min_size=feature_count,
            max_size=feature_count,
        )
    )
    relevant_features = sorted(
        draw(
            st.lists(
                st.integers(min_value=0, max_value=max(0, feature_count - 1)),
                unique=True,
                min_size=0,
                max_size=feature_count,
            )
        )
    )
    return {
        "schema_name": LINEAGE_SCHEMA_NAME,
        "schema_version": LINEAGE_SCHEMA_VERSION,
        "graph": {
            "n_nodes": n_nodes,
            "adjacency": adjacency,
        },
        "assignments": {
            "feature_to_node": feature_to_node,
            "target_to_node": draw(st.integers(min_value=0, max_value=n_nodes - 1)),
            "target_relevant_features": relevant_features,
            "target_relevant_feature_count": len(relevant_features),
            "target_relevant_feature_fraction": (
                float(len(relevant_features)) / float(feature_count) if feature_count > 0 else 0.0
            ),
        },
    }


@st.composite
def _compact_lineage_payload_strategy(draw: st.DrawFn) -> dict[str, Any]:
    n_nodes = draw(_N_NODES_STRATEGY)
    capacity = _upper_triangle_capacity(n_nodes)
    feature_count = draw(_FEATURE_COUNT_STRATEGY)
    feature_to_node = draw(
        st.lists(
            st.integers(min_value=0, max_value=n_nodes - 1),
            min_size=feature_count,
            max_size=feature_count,
        )
    )
    relevant_features = sorted(
        draw(
            st.lists(
                st.integers(min_value=0, max_value=max(0, feature_count - 1)),
                unique=True,
                min_size=0,
                max_size=feature_count,
            )
        )
    )
    return {
        "schema_name": LINEAGE_SCHEMA_NAME,
        "schema_version": LINEAGE_SCHEMA_VERSION_COMPACT,
        "graph": {
            "n_nodes": n_nodes,
            "edge_count": draw(st.integers(min_value=0, max_value=capacity)),
            "adjacency_ref": {
                "encoding": LINEAGE_ADJACENCY_ENCODING,
                "blob_path": draw(_PATH_STRATEGY),
                "index_path": draw(_PATH_STRATEGY),
                "dataset_index": draw(st.integers(min_value=0, max_value=64)),
                "bit_offset": draw(st.integers(min_value=0, max_value=64)).__mul__(8),
                "bit_length": capacity,
                "sha256": draw(_HEX_SHA256_STRATEGY),
            },
        },
        "assignments": {
            "feature_to_node": feature_to_node,
            "target_to_node": draw(st.integers(min_value=0, max_value=n_nodes - 1)),
            "target_relevant_features": relevant_features,
            "target_relevant_feature_count": len(relevant_features),
            "target_relevant_feature_fraction": (
                float(len(relevant_features)) / float(feature_count) if feature_count > 0 else 0.0
            ),
        },
    }


def _valid_lineage_payload() -> dict[str, Any]:
    return {
        "schema_name": LINEAGE_SCHEMA_NAME,
        "schema_version": LINEAGE_SCHEMA_VERSION,
        "graph": {
            "n_nodes": 4,
            "adjacency": [
                [0, 1, 1, 0],
                [0, 0, 1, 1],
                [0, 0, 0, 1],
                [0, 0, 0, 0],
            ],
        },
        "assignments": {
            "feature_to_node": [0, 1, 1, 2, 3],
            "target_to_node": 2,
            "target_relevant_features": [0, 1, 2, 3],
            "target_relevant_feature_count": 4,
            "target_relevant_feature_fraction": 0.8,
        },
    }


def _valid_compact_lineage_payload() -> dict[str, Any]:
    return {
        "schema_name": LINEAGE_SCHEMA_NAME,
        "schema_version": LINEAGE_SCHEMA_VERSION_COMPACT,
        "graph": {
            "n_nodes": 4,
            "edge_count": 4,
            "adjacency_ref": {
                "encoding": LINEAGE_ADJACENCY_ENCODING,
                "blob_path": "../lineage/adjacency.bitpack.bin",
                "index_path": "../lineage/adjacency.index.json",
                "dataset_index": 0,
                "bit_offset": 0,
                "bit_length": 6,
                "sha256": "0" * 64,
            },
        },
        "assignments": {
            "feature_to_node": [0, 1, 1, 2, 3],
            "target_to_node": 2,
            "target_relevant_features": [0, 1, 2, 3],
            "target_relevant_feature_count": 4,
            "target_relevant_feature_fraction": 0.8,
        },
    }


def test_validate_lineage_payload_accepts_valid_payload() -> None:
    validate_lineage_payload(_valid_lineage_payload())


def test_validate_lineage_payload_accepts_valid_compact_payload() -> None:
    validate_lineage_payload(_valid_compact_lineage_payload())


@pytest.mark.parametrize("schema_version", ["1.0.0", "1.1.0", "1.2.0", "1.3.0"])
def test_validate_lineage_payload_rejects_unsupported_legacy_schema_versions(
    schema_version: str,
) -> None:
    payload = _valid_lineage_payload()
    payload["schema_version"] = schema_version

    with pytest.raises(
        LineageValidationError,
        match=rf"lineage\.schema_version: unsupported version '{schema_version}'",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_missing_target_relevant_fields() -> None:
    payload = _valid_lineage_payload()
    del payload["assignments"]["target_relevant_feature_count"]

    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.assignments: missing required key\(s\): target_relevant_feature_count",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_legacy_target_parent_fields_in_current_schema() -> None:
    payload = _valid_lineage_payload()
    payload["assignments"]["target_parent_features"] = [1, 2, 4]

    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.assignments: unknown key\(s\): target_parent_features",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_legacy_target_mode_field() -> None:
    payload = _valid_lineage_payload()
    payload["assignments"]["target_mode"] = "latent_complete_x_conditional"

    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.assignments: unknown key\(s\): target_mode",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_inconsistent_target_relevant_fraction() -> None:
    payload = _valid_lineage_payload()
    payload["assignments"]["target_relevant_feature_fraction"] = 0.2

    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.assignments\.target_relevant_feature_fraction: must equal target_relevant_feature_count / len\(feature_to_node\)",
    ):
        validate_lineage_payload(payload)


@pytest.mark.parametrize(
    ("field_updates", "pattern"),
    [
        (
            {"target_relevant_features": "bad"},
            r"lineage\.assignments\.target_relevant_features: must be a list of feature indices",
        ),
        (
            {"target_relevant_features": [1, "two", 4]},
            (
                r"lineage\.assignments\.target_relevant_features\[1\]: "
                r"must be an integer feature index"
            ),
        ),
        (
            {"target_relevant_features": [1, 5]},
            r"lineage\.assignments\.target_relevant_features\[1\]: must be in range \[0, 4\]",
        ),
        (
            {"target_relevant_features": [1, 1, 4]},
            (
                r"lineage\.assignments\.target_relevant_features\[1\]: "
                r"must be strictly increasing and unique"
            ),
        ),
        (
            {"target_relevant_feature_count": "3"},
            r"lineage\.assignments\.target_relevant_feature_count: must be an integer",
        ),
        (
            {"target_relevant_feature_count": 2},
            r"lineage\.assignments\.target_relevant_feature_count: must equal len\(target_relevant_features\)",
        ),
        (
            {"target_relevant_feature_fraction": "0.8"},
            r"lineage\.assignments\.target_relevant_feature_fraction: must be a finite ratio in \[0, 1\]",
        ),
        (
            {"target_relevant_feature_fraction": 1.1},
            r"lineage\.assignments\.target_relevant_feature_fraction: must be a finite ratio in \[0, 1\]",
        ),
    ],
)
def test_validate_lineage_payload_rejects_invalid_target_relevance_metadata(
    field_updates: dict[str, object],
    pattern: str,
) -> None:
    payload = _valid_lineage_payload()
    payload["assignments"].update(field_updates)

    with pytest.raises(LineageValidationError, match=pattern):
        validate_lineage_payload(payload)


def test_validate_metadata_lineage_accepts_absent_payload_when_optional() -> None:
    validate_metadata_lineage({"seed": 1}, required=False)


def test_validate_metadata_lineage_rejects_missing_payload_when_required() -> None:
    with pytest.raises(LineageValidationError, match=r"metadata\.lineage: is required"):
        validate_metadata_lineage({"seed": 1}, required=True)


def test_validate_lineage_payload_rejects_wrong_schema_name() -> None:
    payload = _valid_lineage_payload()
    payload["schema_name"] = "other.schema"
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.schema_name: must equal 'dagzoo\.dag_lineage'",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_unknown_root_keys() -> None:
    payload = _valid_lineage_payload()
    payload["extra"] = True
    with pytest.raises(LineageValidationError, match=r"lineage: unknown key\(s\): extra"):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_non_square_adjacency() -> None:
    payload = _valid_lineage_payload()
    graph = deepcopy(payload["graph"])
    assert isinstance(graph, dict)
    graph["adjacency"] = [[0, 1], [0, 0]]
    payload["graph"] = graph
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.graph\.adjacency: must have 4 rows \(got 2\)",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_non_binary_adjacency_values() -> None:
    payload = _valid_lineage_payload()
    graph = deepcopy(payload["graph"])
    assert isinstance(graph, dict)
    adjacency = deepcopy(graph["adjacency"])
    assert isinstance(adjacency, list)
    adjacency[0][1] = 2
    graph["adjacency"] = adjacency
    payload["graph"] = graph
    with pytest.raises(
        LineageValidationError, match=r"lineage\.graph\.adjacency\[0\]\[1\]: must be 0 or 1"
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_bool_adjacency_values() -> None:
    payload = _valid_lineage_payload()
    graph = deepcopy(payload["graph"])
    assert isinstance(graph, dict)
    adjacency = deepcopy(graph["adjacency"])
    assert isinstance(adjacency, list)
    adjacency[0][1] = True
    graph["adjacency"] = adjacency
    payload["graph"] = graph
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.graph\.adjacency\[0\]\[1\]: must be integer 0 or 1",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_diagonal_edges() -> None:
    payload = _valid_lineage_payload()
    graph = deepcopy(payload["graph"])
    assert isinstance(graph, dict)
    adjacency = deepcopy(graph["adjacency"])
    assert isinstance(adjacency, list)
    adjacency[2][2] = 1
    graph["adjacency"] = adjacency
    payload["graph"] = graph
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.graph\.adjacency\[2\]\[2\]: must be 0 on the diagonal",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_lower_triangle_edges() -> None:
    payload = _valid_lineage_payload()
    graph = deepcopy(payload["graph"])
    assert isinstance(graph, dict)
    adjacency = deepcopy(graph["adjacency"])
    assert isinstance(adjacency, list)
    adjacency[3][1] = 1
    graph["adjacency"] = adjacency
    payload["graph"] = graph
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.graph\.adjacency\[3\]\[1\]: must be 0 for upper-triangular DAG encoding",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_feature_assignment_out_of_range() -> None:
    payload = _valid_lineage_payload()
    assignments = deepcopy(payload["assignments"])
    assert isinstance(assignments, dict)
    assignments["feature_to_node"] = [0, 1, 4]
    payload["assignments"] = assignments
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.assignments\.feature_to_node\[2\]: must be in range \[0, 3\]",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_target_assignment_out_of_range() -> None:
    payload = _valid_lineage_payload()
    assignments = deepcopy(payload["assignments"])
    assert isinstance(assignments, dict)
    assignments["target_to_node"] = -1
    payload["assignments"] = assignments
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.assignments\.target_to_node: must be in range \[0, 3\]",
    ):
        validate_lineage_payload(payload)


def test_validate_metadata_lineage_rejects_non_object_payload() -> None:
    with pytest.raises(LineageValidationError, match=r"metadata\.lineage: must be an object"):
        validate_metadata_lineage({"lineage": 42}, required=False)


def test_validate_lineage_payload_rejects_unknown_compact_encoding() -> None:
    payload = _valid_compact_lineage_payload()
    graph = deepcopy(payload["graph"])
    assert isinstance(graph, dict)
    adjacency_ref = deepcopy(graph["adjacency_ref"])
    assert isinstance(adjacency_ref, dict)
    adjacency_ref["encoding"] = "other"
    graph["adjacency_ref"] = adjacency_ref
    payload["graph"] = graph
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.graph\.adjacency_ref\.encoding: must equal 'upper_triangle_bitpack_v1'",
    ):
        validate_lineage_payload(payload)


def test_validate_lineage_payload_rejects_compact_bit_length_mismatch() -> None:
    payload = _valid_compact_lineage_payload()
    graph = deepcopy(payload["graph"])
    assert isinstance(graph, dict)
    adjacency_ref = deepcopy(graph["adjacency_ref"])
    assert isinstance(adjacency_ref, dict)
    adjacency_ref["bit_length"] = 5
    graph["adjacency_ref"] = adjacency_ref
    payload["graph"] = graph
    with pytest.raises(
        LineageValidationError,
        match=r"lineage\.graph\.adjacency_ref\.bit_length: must equal 6 for n_nodes=4",
    ):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_dense_lineage_payload_strategy())
def test_validate_lineage_payload_accepts_generated_dense_payloads_hypothesis(
    payload: dict[str, Any],
) -> None:
    validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_compact_lineage_payload_strategy())
def test_validate_lineage_payload_accepts_generated_compact_payloads_hypothesis(
    payload: dict[str, Any],
) -> None:
    validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_dense_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_diagonal_edge_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    payload["graph"]["adjacency"][0][0] = 1
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_dense_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_lower_triangle_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    payload["graph"]["adjacency"][1][0] = 1
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_dense_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_nonbinary_adjacency_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    payload["graph"]["adjacency"][0][1] = 2
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_dense_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_feature_assignment_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    n_nodes = int(payload["graph"]["n_nodes"])
    payload["assignments"]["feature_to_node"] = [n_nodes]
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_dense_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_target_assignment_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    payload["assignments"]["target_to_node"] = -1
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_compact_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_compact_encoding_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    payload["graph"]["adjacency_ref"]["encoding"] = "other"
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_compact_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_compact_bit_length_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    payload["graph"]["adjacency_ref"]["bit_length"] += 1
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_compact_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_compact_bit_offset_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    payload["graph"]["adjacency_ref"]["bit_offset"] += 4
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_compact_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_compact_sha256_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    payload["graph"]["adjacency_ref"]["sha256"] = "g" * 64
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)


@settings(max_examples=100, deadline=None)
@given(payload=_compact_lineage_payload_strategy())
def test_validate_lineage_payload_rejects_generated_compact_edge_count_mutations_hypothesis(
    payload: dict[str, Any],
) -> None:
    n_nodes = int(payload["graph"]["n_nodes"])
    payload["graph"]["edge_count"] = _upper_triangle_capacity(n_nodes) + 1
    with pytest.raises(LineageValidationError):
        validate_lineage_payload(payload)



"""Coverage-summary comparison helpers for diversity audits."""

from __future__ import annotations

from statistics import median
from typing import Any

from dagzoo.math import coerce_optional_finite_float as _coerce_optional_finite_float

CORE_DIVERSITY_METRICS: tuple[str, ...] = (
    "linearity_proxy",
    "nonlinearity_proxy",
    "wins_ratio_proxy",
    "pearson_abs_mean",
    "pearson_abs_max",
    "snr_proxy_db",
    "class_entropy",
    "majority_minority_ratio",
    "categorical_ratio",
    "cat_cardinality_mean",
    "graph_edge_density",
)


def validate_diversity_thresholds(
    *,
    warn_threshold_pct: float,
    fail_threshold_pct: float,
) -> tuple[float, float]:
    """Validate diversity threshold values and ordering."""

    warn_value = _coerce_optional_finite_float(warn_threshold_pct)
    fail_value = _coerce_optional_finite_float(fail_threshold_pct)
    if warn_value is None or fail_value is None:
        raise ValueError("warn_threshold_pct and fail_threshold_pct must be finite values.")
    if warn_value < 0.0 or fail_value < 0.0:
        raise ValueError("warn_threshold_pct and fail_threshold_pct must be >= 0.")
    if warn_value > fail_value:
        raise ValueError("warn_threshold_pct must be <= fail_threshold_pct.")
    return float(warn_value), float(fail_value)


def _metric_payload(summary: dict[str, Any], metric: str) -> dict[str, Any] | None:
    metrics = summary.get("metrics")
    if not isinstance(metrics, dict):
        return None
    payload = metrics.get(metric)
    if not isinstance(payload, dict):
        return None
    return payload


def _metric_mean(summary: dict[str, Any], metric: str) -> float | None:
    payload = _metric_payload(summary, metric)
    if payload is None:
        return None
    return _coerce_optional_finite_float(payload.get("mean"))


def _metric_quantile(summary: dict[str, Any], metric: str, key: str) -> float | None:
    payload = _metric_payload(summary, metric)
    if payload is None:
        return None
    quantiles = payload.get("quantiles")
    if not isinstance(quantiles, dict):
        return None
    return _coerce_optional_finite_float(quantiles.get(key))


def _delta_pct(current: float | None, baseline: float | None) -> float | None:
    """Return percentage delta vs baseline when both values are comparable."""

    if current is None or baseline is None:
        return None
    if baseline == 0.0:
        return 0.0 if current == 0.0 else None
    return float(((current - baseline) / abs(baseline)) * 100.0)


def metric_shift_pct(
    *,
    baseline_summary: dict[str, Any],
    variant_summary: dict[str, Any],
    metric: str,
) -> float | None:
    """Compute the normalized percent shift for one diversity metric."""

    baseline_mean = _metric_mean(baseline_summary, metric)
    variant_mean = _metric_mean(variant_summary, metric)
    baseline_p50 = _metric_quantile(baseline_summary, metric, "p50")
    variant_p50 = _metric_quantile(variant_summary, metric, "p50")
    baseline_p25 = _metric_quantile(baseline_summary, metric, "p25")
    baseline_p75 = _metric_quantile(baseline_summary, metric, "p75")
    if (
        baseline_mean is None
        or variant_mean is None
        or baseline_p50 is None
        or variant_p50 is None
        or baseline_p25 is None
        or baseline_p75 is None
    ):
        return None
    iqr = abs(float(baseline_p75) - float(baseline_p25))
    scale = max(abs(float(baseline_mean)), iqr, 1e-6)
    weighted_shift = (0.6 * abs(float(variant_mean) - float(baseline_mean)) / scale) + (
        0.4 * abs(float(variant_p50) - float(baseline_p50)) / scale
    )
    return float(weighted_shift * 100.0)


def classify_diversity_status(
    composite_shift_pct: float | None,
    *,
    warn_threshold_pct: float,
    fail_threshold_pct: float,
) -> str:
    """Classify one comparison severity from the composite shift."""

    warn_threshold_pct, fail_threshold_pct = validate_diversity_thresholds(
        warn_threshold_pct=warn_threshold_pct,
        fail_threshold_pct=fail_threshold_pct,
    )
    if composite_shift_pct is None:
        return "insufficient_metrics"
    if float(composite_shift_pct) >= float(fail_threshold_pct):
        return "fail"
    if float(composite_shift_pct) >= float(warn_threshold_pct):
        return "warn"
    return "pass"


def compare_coverage_summaries(
    *,
    baseline_summary: dict[str, Any],
    variant_summary: dict[str, Any],
    warn_threshold_pct: float,
    fail_threshold_pct: float,
) -> dict[str, Any]:
    """Compare one variant coverage summary against a baseline summary."""

    metric_shift_map: dict[str, float] = {}
    for metric in CORE_DIVERSITY_METRICS:
        shift = metric_shift_pct(
            baseline_summary=baseline_summary,
            variant_summary=variant_summary,
            metric=metric,
        )
        if shift is not None:
            metric_shift_map[metric] = float(shift)
    composite_shift_pct = float(median(metric_shift_map.values())) if metric_shift_map else None
    return {
        "diversity_metric_shift_pct": metric_shift_map,
        "diversity_composite_shift_pct": composite_shift_pct,
        "diversity_status": classify_diversity_status(
            composite_shift_pct,
            warn_threshold_pct=warn_threshold_pct,
            fail_threshold_pct=fail_threshold_pct,
        ),
    }


def build_comparison_record(
    *,
    baseline_entry: dict[str, Any],
    variant_entry: dict[str, Any],
    warn_threshold_pct: float,
    fail_threshold_pct: float,
) -> dict[str, Any]:
    """Build one report comparison record for a variant entry."""

    comparison = compare_coverage_summaries(
        baseline_summary=baseline_entry["coverage_summary"],
        variant_summary=variant_entry["coverage_summary"],
        warn_threshold_pct=warn_threshold_pct,
        fail_threshold_pct=fail_threshold_pct,
    )
    comparison["variant_label"] = variant_entry["label"]
    comparison["datasets_per_minute_delta_pct"] = _delta_pct(
        variant_entry.get("datasets_per_minute"),
        baseline_entry.get("datasets_per_minute"),
    )
    comparison["filter_accepted_datasets_per_minute_delta_pct"] = _delta_pct(
        variant_entry.get("filter_accepted_datasets_per_minute"),
        baseline_entry.get("filter_accepted_datasets_per_minute"),
    )
    return comparison


def summarize_comparison_status(comparisons: list[dict[str, Any]]) -> str:
    """Collapse per-variant severities into one audit-level status."""

    statuses = {str(item.get("diversity_status", "insufficient_metrics")) for item in comparisons}
    if "fail" in statuses:
        return "fail"
    if "insufficient_metrics" in statuses:
        return "insufficient_metrics"
    if "warn" in statuses:
        return "warn"
    if "pass" in statuses:
        return "pass"
    return "insufficient_metrics"



"""Typed configuration models and file loading."""

from .constants import (
    DATASET_ROWS_MAX_TOTAL,
    DATASET_ROWS_MIN_TOTAL,
    LAYOUT_MODE_FIXED,
    LAYOUT_MODE_HETEROGENEOUS,
    MAX_SUPPORTED_CLASS_COUNT,
    MISSINGNESS_MECHANISM_MAR,
    MISSINGNESS_MECHANISM_MCAR,
    MISSINGNESS_MECHANISM_MNAR,
    MISSINGNESS_MECHANISM_NONE,
    NOISE_FAMILY_GAUSSIAN,
    NOISE_FAMILY_LAPLACE,
    NOISE_FAMILY_MIXTURE,
    NOISE_FAMILY_STUDENT_T,
    NOISE_MIXTURE_COMPONENT_GAUSSIAN,
    NOISE_MIXTURE_COMPONENT_LAPLACE,
    NOISE_MIXTURE_COMPONENT_STUDENT_T,
    SHIFT_MODE_CUSTOM,
    SHIFT_MODE_GRAPH_DRIFT,
    SHIFT_MODE_MECHANISM_DRIFT,
    SHIFT_MODE_MIXED,
    SHIFT_MODE_NOISE_DRIFT,
    SHIFT_MODE_OFF,
    LayoutMode,
    MechanismFamily,
    MissingnessMechanism,
    NoiseFamily,
    NoiseMixtureComponent,
    RowsMode,
    ShiftMode,
)
from .io import clone_generator_config, load_packaged_generator_config
from .models import (
    BenchmarkConfig,
    DatasetConfig,
    DiagnosticsConfig,
    FilterConfig,
    GeneratorConfig,
    GraphConfig,
    MechanismConfig,
    NoiseConfig,
    OutputConfig,
    RuntimeConfig,
    ShiftConfig,
    SteeringConfig,
    SteeringStageConfig,
    StressConfig,
    effective_config_payload,
    materialize_stress_profile,
    normalize_filter_config,
    stress_profile_definition,
)
from .normalization import (
    normalize_layout_mode,
    normalize_missing_mechanism,
    normalize_noise_family,
    normalize_shift_mode,
)
from .rows import (
    DatasetRowsSpec,
    dataset_rows_bounds,
    dataset_rows_is_variable,
    normalize_dataset_rows,
    resolve_dataset_total_rows,
    validate_class_split_feasibility,
)

__all__ = [
    "DATASET_ROWS_MAX_TOTAL",
    "DATASET_ROWS_MIN_TOTAL",
    "LAYOUT_MODE_FIXED",
    "LAYOUT_MODE_HETEROGENEOUS",
    "MAX_SUPPORTED_CLASS_COUNT",
    "MISSINGNESS_MECHANISM_MAR",
    "MISSINGNESS_MECHANISM_MCAR",
    "MISSINGNESS_MECHANISM_MNAR",
    "MISSINGNESS_MECHANISM_NONE",
    "NOISE_FAMILY_GAUSSIAN",
    "NOISE_FAMILY_LAPLACE",
    "NOISE_FAMILY_MIXTURE",
    "NOISE_FAMILY_STUDENT_T",
    "NOISE_MIXTURE_COMPONENT_GAUSSIAN",
    "NOISE_MIXTURE_COMPONENT_LAPLACE",
    "NOISE_MIXTURE_COMPONENT_STUDENT_T",
    "SHIFT_MODE_CUSTOM",
    "SHIFT_MODE_GRAPH_DRIFT",
    "SHIFT_MODE_MECHANISM_DRIFT",
    "SHIFT_MODE_MIXED",
    "SHIFT_MODE_NOISE_DRIFT",
    "SHIFT_MODE_OFF",
    "BenchmarkConfig",
    "DatasetConfig",
    "DatasetRowsSpec",
    "DiagnosticsConfig",
    "FilterConfig",
    "GeneratorConfig",
    "GraphConfig",
    "LayoutMode",
    "MechanismConfig",
    "MechanismFamily",
    "MissingnessMechanism",
    "NoiseConfig",
    "NoiseFamily",
    "NoiseMixtureComponent",
    "OutputConfig",
    "RowsMode",
    "RuntimeConfig",
    "ShiftConfig",
    "ShiftMode",
    "SteeringConfig",
    "SteeringStageConfig",
    "StressConfig",
    "clone_generator_config",
    "dataset_rows_bounds",
    "dataset_rows_is_variable",
    "effective_config_payload",
    "load_packaged_generator_config",
    "materialize_stress_profile",
    "normalize_dataset_rows",
    "normalize_filter_config",
    "normalize_layout_mode",
    "normalize_missing_mechanism",
    "normalize_noise_family",
    "normalize_shift_mode",
    "resolve_dataset_total_rows",
    "stress_profile_definition",
    "validate_class_split_feasibility",
]



"""Shared config state predicates used across runtime and benchmark code."""

from __future__ import annotations

from dagzoo.config import (
    MISSINGNESS_MECHANISM_NONE,
    NOISE_FAMILY_GAUSSIAN,
    GeneratorConfig,
)


def missingness_enabled(config: GeneratorConfig) -> bool:
    """Return whether missingness injection can mutate emitted tensors."""

    return bool(
        float(config.dataset.missing_rate) > 0.0
        and str(config.dataset.missing_mechanism).strip().lower() != MISSINGNESS_MECHANISM_NONE
    )


def shift_enabled(config: GeneratorConfig) -> bool:
    """Return whether shift controls are enabled in config."""

    return bool(config.shift.enabled)


def non_gaussian_noise_enabled(config: GeneratorConfig) -> bool:
    """Return whether non-Gaussian noise controls are enabled in config."""

    return str(config.noise.family).strip().lower() != NOISE_FAMILY_GAUSSIAN



from __future__ import annotations

import shutil
import subprocess
from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

REPO_ROOT = Path(__file__).resolve().parents[2]
SRC_ROOT = REPO_ROOT / "src" / "dagzoo"
DOCS_DEP_MAP_PATH = REPO_ROOT / "docs" / "development" / "module-dependency-map.md"


@dataclass(frozen=True)
class CommandSpec:
    label: str
    argv: tuple[str, ...]


class DevToolError(RuntimeError):
    """Raised for developer tooling failures with user-facing messages."""


def repo_relative(path: Path) -> str:
    try:
        return path.relative_to(REPO_ROOT).as_posix()
    except ValueError:
        return path.as_posix()


def python_tool(tool_name: str) -> str:
    candidate = REPO_ROOT / ".venv" / "bin" / tool_name
    if candidate.exists():
        return str(candidate)
    return tool_name


def venv_python() -> Path:
    return REPO_ROOT / ".venv" / "bin" / "python"


def run_command(command: CommandSpec) -> None:
    result = subprocess.run(command.argv, cwd=REPO_ROOT, check=False)
    if result.returncode != 0:
        raise DevToolError(f"{command.label} failed with exit code {result.returncode}.")


def format_command(argv: tuple[str, ...]) -> str:
    return " ".join(argv)


def tool_exists(tool_name: str) -> bool:
    return shutil.which(tool_name) is not None


def normalize_files(files: Sequence[str]) -> tuple[str, ...]:
    normalized: list[str] = []
    for file_str in files:
        if not file_str:
            continue
        path = Path(file_str)
        if path.is_absolute():
            normalized.append(repo_relative(path))
        else:
            normalized.append(path.as_posix())
    return tuple(normalized)


def run_git_capture(*args: str) -> str:
    result = subprocess.run(
        ("git", *args),
        cwd=REPO_ROOT,
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        detail = result.stderr.strip() or result.stdout.strip() or "git command failed."
        raise RuntimeError(f"git {' '.join(args)} failed: {detail}")
    return result.stdout


def run_git_lines(*args: str) -> tuple[str, ...]:
    return tuple(line.strip() for line in run_git_capture(*args).splitlines() if line.strip())


def git_common_dir() -> Path:
    return Path(run_git_capture("rev-parse", "--path-format=absolute", "--git-common-dir").strip())


def git_hook_path(hook_name: str) -> Path:
    return git_common_dir() / "hooks" / hook_name



"""Fixed-layout run preparation helpers."""

from __future__ import annotations

from dagzoo.config import (
    DatasetRowsSpec,
    GeneratorConfig,
    clone_generator_config,
    dataset_rows_is_variable,
    materialize_stress_profile,
    resolve_dataset_total_rows,
)
from dagzoo.core.generation_context import _resolve_device, _resolve_run_seed
from dagzoo.rng import KeyedRng

from .metadata import _FixedLayoutPlan

_FIXED_LAYOUT_TARGET_CELLS = 4_000_000


def normalize_fixed_layout_target_cells(target_cells: int | None) -> int:
    """Return the effective fixed-layout target cell budget for a runtime value."""

    if target_cells is None:
        return int(_FIXED_LAYOUT_TARGET_CELLS)
    return int(target_cells)


def _validate_fixed_layout_rows_mode(config: GeneratorConfig) -> None:
    if dataset_rows_is_variable(config.dataset.rows):
        raise ValueError(
            "Fixed-layout generation requires a fixed split size; variable dataset.rows "
            "modes (range/choices) are not supported."
        )


def _effective_fixed_layout_target_cells(config: GeneratorConfig) -> int:
    """Return the configured fixed-layout auto-batch target cell budget."""

    return normalize_fixed_layout_target_cells(config.runtime.fixed_layout_target_cells)


def _resolve_fixed_layout_batch_size(
    plan: _FixedLayoutPlan,
    *,
    num_datasets: int,
    batch_size: int | None,
    target_cells: int | None = None,
    batch_size_cap: int | None = None,
) -> int:
    effective_cap = None if batch_size_cap is None else max(1, int(batch_size_cap))
    if batch_size is not None:
        resolved = max(1, min(int(batch_size), int(num_datasets)))
        return resolved if effective_cap is None else min(resolved, effective_cap)
    per_dataset_cells = max(
        1, int(plan.n_train + plan.n_test) * max(1, int(plan.layout.n_features))
    )
    auto_batch = max(1, int(target_cells or _FIXED_LAYOUT_TARGET_CELLS) // per_dataset_cells)
    resolved = max(1, min(int(num_datasets), int(auto_batch)))
    return resolved if effective_cap is None else min(resolved, effective_cap)


def realize_generation_config_for_run(
    config: GeneratorConfig,
    *,
    seed: int | None = None,
    device: str | None = None,
) -> tuple[GeneratorConfig, int, str, str]:
    """Resolve one canonical single-run config with rows fixed for the full run."""

    run_seed = _resolve_run_seed(config, seed)
    requested_device = (device or config.runtime.device or "auto").lower()
    resolved_device = _resolve_device(config, device)

    if config.stress.profile is not None:
        config.validate_generation_constraints()
        realized = materialize_stress_profile(
            config,
            revalidate=False,
            clear_selector=True,
        )
    else:
        realized = clone_generator_config(config, revalidate=False)
    rows_seed = KeyedRng(run_seed).child_seed("rows")
    total_rows = resolve_dataset_total_rows(realized.dataset.rows, dataset_seed=rows_seed)
    if total_rows is not None:
        n_test = int(realized.dataset.n_test)
        n_train = int(total_rows) - n_test
        if n_train <= 0:
            raise ValueError(
                "Resolved rows split is invalid: total rows must be > dataset.n_test "
                f"(total_rows={int(total_rows)}, n_test={n_test})."
            )
        realized.dataset.n_train = int(n_train)
        realized.dataset.rows = DatasetRowsSpec(mode="fixed", value=int(total_rows))

    return realized, int(run_seed), str(requested_device), str(resolved_device)


__all__ = [
    "_effective_fixed_layout_target_cells",
    "_resolve_fixed_layout_batch_size",
    "_validate_fixed_layout_rows_mode",
    "normalize_fixed_layout_target_cells",
    "realize_generation_config_for_run",
]



"""Rewritten baseline-vs-variant diversity audit runner."""

from __future__ import annotations

import datetime as dt
from typing import Any

from dagzoo.bench.corpus_probe import (
    CorpusProbeResult,
    build_corpus_probe_coverage_config,
    resolve_corpus_probe_counts,
    run_corpus_probe,
)
from dagzoo.config import GeneratorConfig
from dagzoo.recipes import parse_recipe_reference, serialize_config_reference

from .compare import (
    CORE_DIVERSITY_METRICS,
    build_comparison_record,
    summarize_comparison_status,
)


def _probe_entry(result: CorpusProbeResult) -> dict[str, Any]:
    """Convert one probe dataclass into the persisted report entry shape."""

    mechanism_family_summary = None
    if isinstance(result.coverage_summary, dict):
        payload = result.coverage_summary.get("mechanism_family_summary")
        if isinstance(payload, dict):
            mechanism_family_summary = payload
    return {
        "label": result.label,
        "config_path": result.config_path,
        "suite": result.suite,
        "num_datasets": int(result.num_datasets),
        "warmup_datasets": int(result.warmup_datasets),
        "requested_device": result.requested_device,
        "resolved_device": result.resolved_device,
        "resolved_config": result.resolved_config,
        "datasets_per_minute": float(result.datasets_per_minute),
        "filter_datasets_per_minute": result.filter_datasets_per_minute,
        "filter_accepted_datasets_per_minute": result.filter_accepted_datasets_per_minute,
        "filter_accepted_datasets_measured": int(result.filter_accepted_datasets_measured),
        "filter_rejected_datasets_measured": int(result.filter_rejected_datasets_measured),
        "filter_acceptance_rate_dataset_level": result.filter_acceptance_rate_dataset_level,
        "filter_rejection_rate_dataset_level": result.filter_rejection_rate_dataset_level,
        "coverage_summary": result.coverage_summary,
        "mechanism_family_summary": mechanism_family_summary,
        "filter_summary": result.filter_summary,
    }


def _build_variant_labels(variant_config_paths: list[str]) -> list[str]:
    """Derive stable labels from variant config paths while avoiding duplicates."""

    seen: dict[str, int] = {}
    labels: list[str] = []
    for idx, path in enumerate(variant_config_paths, start=1):
        recipe_name = parse_recipe_reference(path)
        if recipe_name is not None:
            stem = recipe_name
        else:
            stem = path.rsplit("/", 1)[-1].rsplit(".", 1)[0] if path else f"variant_{idx}"
        count = seen.get(stem, 0)
        seen[stem] = count + 1
        labels.append(stem if count == 0 else f"{stem}_{count + 1}")
    return labels


def run_effective_diversity_audit(
    *,
    baseline_config: GeneratorConfig,
    baseline_config_path: str,
    variant_configs: list[GeneratorConfig],
    variant_config_paths: list[str],
    variant_labels: list[str] | None = None,
    suite: str,
    num_datasets: int | None,
    warmup: int | None,
    device: str | None,
    warn_threshold_pct: float,
    fail_threshold_pct: float,
) -> dict[str, Any]:
    """Run the rewritten diversity audit against baseline and variant configs."""

    if len(variant_configs) != len(variant_config_paths):
        raise ValueError("variant_configs and variant_config_paths must have matching lengths.")
    if variant_labels is not None and len(variant_labels) != len(variant_configs):
        raise ValueError("variant_labels and variant_configs must have matching lengths.")

    probe_seed = int(baseline_config.seed)
    probe_coverage_config = build_corpus_probe_coverage_config(baseline_config)
    probe_num_datasets, probe_warmup_datasets = resolve_corpus_probe_counts(
        baseline_config,
        suite=suite,
        num_datasets_override=num_datasets,
        warmup_override=warmup,
    )
    baseline_probe = run_corpus_probe(
        baseline_config,
        label="baseline",
        config_path=serialize_config_reference(baseline_config_path),
        suite=suite,
        num_datasets=probe_num_datasets,
        warmup=probe_warmup_datasets,
        device=device,
        probe_seed=probe_seed,
        coverage_config=probe_coverage_config,
    )
    baseline_entry = _probe_entry(baseline_probe)

    variants: list[dict[str, Any]] = []
    comparisons: list[dict[str, Any]] = []
    for label, config_path, variant_config in zip(
        variant_labels
        if variant_labels is not None
        else _build_variant_labels(variant_config_paths),
        variant_config_paths,
        variant_configs,
        strict=True,
    ):
        variant_probe = run_corpus_probe(
            variant_config,
            label=label,
            config_path=serialize_config_reference(config_path),
            suite=suite,
            num_datasets=probe_num_datasets,
            warmup=probe_warmup_datasets,
            device=device,
            probe_seed=probe_seed,
            coverage_config=probe_coverage_config,
        )
        variant_entry = _probe_entry(variant_probe)
        variants.append(variant_entry)
        comparisons.append(
            build_comparison_record(
                baseline_entry=baseline_entry,
                variant_entry=variant_entry,
                warn_threshold_pct=warn_threshold_pct,
                fail_threshold_pct=fail_threshold_pct,
            )
        )

    status_counts = {
        "pass": 0,
        "warn": 0,
        "fail": 0,
        "insufficient_metrics": 0,
    }
    for comparison in comparisons:
        status = str(comparison.get("diversity_status", "insufficient_metrics"))
        status_counts[status] = status_counts.get(status, 0) + 1

    return {
        "schema_name": "dagzoo_diversity_audit_report",
        "schema_version": 1,
        "generated_at": dt.datetime.now(dt.UTC).isoformat(),
        "baseline": baseline_entry,
        "variants": variants,
        "comparisons": comparisons,
        "summary": {
            "overall_status": summarize_comparison_status(comparisons),
            "warn_threshold_pct": float(warn_threshold_pct),
            "fail_threshold_pct": float(fail_threshold_pct),
            "num_variants": len(variants),
            "probe_num_datasets": int(probe_num_datasets),
            "probe_warmup_datasets": int(probe_warmup_datasets),
            "status_counts": status_counts,
            "core_metrics": list(CORE_DIVERSITY_METRICS),
        },
    }



"""Dataset-level diagnostics metrics extraction."""

from .coverage import (
    CoverageAggregationConfig,
    CoverageAggregator,
    write_coverage_summary_json,
    write_coverage_summary_markdown,
)
from .metrics import extract_dataset_metrics, extract_metrics_batch
from .types import DatasetMetrics

__all__ = [
    "CoverageAggregationConfig",
    "CoverageAggregator",
    "DatasetMetrics",
    "extract_dataset_metrics",
    "extract_metrics_batch",
    "write_coverage_summary_json",
    "write_coverage_summary_markdown",
]



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import sys
from pathlib import Path

SCRIPT_DIR = Path(__file__).resolve().parent
if str(SCRIPT_DIR) not in sys.path:
    sys.path.insert(0, str(SCRIPT_DIR))

from devlib.bootstrap import bootstrap_environment  # noqa: E402
from devlib.common import DevToolError  # noqa: E402
from devlib.contract import evaluate_release_contract, render_contract_result  # noqa: E402
from devlib.impact import (  # noqa: E402
    build_impact_report,
    detect_changed_files,
)
from devlib.impact import render_json as render_impact_json  # noqa: E402
from devlib.impact import render_text as render_impact_text  # noqa: E402
from devlib.review import (  # noqa: E402
    DEFAULT_BASE_REF,
    build_review_base_result,
    render_review_base_report,
)


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="./scripts/dev",
        description="Repo-local developer tooling for dagzoo.",
        epilog=(
            "Canonical local flow:\n"
            "  ./scripts/dev bootstrap\n"
            "  ./.venv/bin/nox -s quick\n"
            "  ./.venv/bin/nox -s bench_smoke\n"
            "  ./scripts/dev review-base\n"
            "  ./scripts/dev impact"
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    subparsers = parser.add_subparsers(dest="command", required=True)

    subparsers.add_parser("bootstrap")

    impact_parser = subparsers.add_parser("impact")
    _add_change_source_args(impact_parser)
    impact_parser.add_argument("--format", choices=("text", "json"), default="text")

    contract_parser = subparsers.add_parser("contract")
    _add_change_source_args(contract_parser)
    contract_parser.add_argument("--strict", action="store_true")

    review_parser = subparsers.add_parser("review-base")
    review_parser.add_argument("--base-ref", default=DEFAULT_BASE_REF)

    return parser


def _add_change_source_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--source", choices=("working-tree", "staged", "base"), default="working-tree"
    )
    parser.add_argument("--base")
    parser.add_argument("--files", nargs="*")


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    try:
        if args.command == "bootstrap":
            print(bootstrap_environment(), end="")
            return 0

        if args.command == "impact":
            changed_files = detect_changed_files(
                source=args.source,
                base=args.base,
                files=args.files,
            )
            report = build_impact_report(changed_files)
            rendered = (
                render_impact_json(report) if args.format == "json" else render_impact_text(report)
            )
            print(rendered, end="")
            return 0

        if args.command == "contract":
            changed_files = detect_changed_files(
                source=args.source,
                base=args.base,
                files=args.files,
            )
            report = build_impact_report(changed_files)
            result = evaluate_release_contract(report, strict=args.strict)
            print(render_contract_result(result), end="")
            return 0 if result.ok else 1

        if args.command == "review-base":
            result = build_review_base_result(args.base_ref)
            print(render_review_base_report(result), end="")
            return 0 if result.contract.ok else 1
    except DevToolError as exc:
        print(str(exc), file=sys.stderr)
        return 1
    except RuntimeError as exc:
        print(str(exc), file=sys.stderr)
        return 1
    parser.error("unreachable")
    return 2


if __name__ == "__main__":
    raise SystemExit(main())



"""Random points mechanism."""

import torch

from dagzoo.core.execution_semantics import sample_root_source_plan
from dagzoo.core.fixed_layout.batched import (
    FixedLayoutBatchRng,
    _sample_random_points_batch,
    apply_function_plan_batch,
)
from dagzoo.core.layout_types import MechanismFamily
from dagzoo.rng import keyed_rng_from_generator
from dagzoo.sampling.noise import NoiseSamplingSpec


def sample_random_points(
    n_rows: int,
    dim: int,
    generator: torch.Generator,
    device: str,
    *,
    mechanism_logit_tilt: float = 0.0,
    function_family_mix: dict[MechanismFamily, float] | None = None,
    noise_sigma_multiplier: float = 1.0,
    noise_spec: NoiseSamplingSpec | None = None,
) -> torch.Tensor:
    """Sample one typed random-points source and execute it in torch."""
    if n_rows <= 0 or dim <= 0:
        raise ValueError(f"n_rows and dim must be > 0. Got n_rows={n_rows}, dim={dim}")

    root = keyed_rng_from_generator(generator, "sample_random_points")
    source = sample_root_source_plan(
        keyed_rng=root.keyed("plan"),
        out_dim=dim,
        mechanism_logit_tilt=mechanism_logit_tilt,
        function_family_mix=function_family_mix,
        device=str(generator.device),
    )
    rng = FixedLayoutBatchRng.from_keyed_rng(
        root.keyed("execution", "source"),
        batch_size=1,
        device=device,
    )
    base = _sample_random_points_batch(
        rng.keyed("base"),
        n_rows=n_rows,
        dim=dim,
        base_kind=source.base_kind,
        noise_sigma_multiplier=noise_sigma_multiplier,
        noise_spec=noise_spec,
    )
    out = apply_function_plan_batch(
        base,
        rng.keyed("function"),
        source.function,
        out_dim=dim,
        noise_sigma_multiplier=noise_sigma_multiplier,
        noise_spec=noise_spec,
    )
    return out.squeeze(0)



from __future__ import annotations

import json

import pytest
import yaml
from conftest import load_repo_config, write_config

from dagzoo.bench.constants import SMOKE_N_TEST_CAP, SMOKE_N_TRAIN_CAP
from dagzoo.cli.commands.benchmark import _print_preset_result_line
from dagzoo.cli.effective_config import effective_config_payload_yaml, effective_config_yaml
from dagzoo.cli.entrypoint import main
from dagzoo.cli.parsing import DEVICE_CHOICES, HARDWARE_POLICY_CHOICES
from dagzoo.config import effective_config_payload


def test_cli_parser_choice_constants_are_stable() -> None:
    assert DEVICE_CHOICES == ("auto", "cpu", "cuda", "mps")
    assert "none" in HARDWARE_POLICY_CHOICES


def test_benchmark_cli_writes_json_with_scenarios(tmp_path) -> None:
    out = tmp_path / "summary.json"
    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "2",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--json-out",
            str(out),
        ]
    )
    assert code == 0
    payload = json.loads(out.read_text(encoding="utf-8"))
    profile = payload["preset_results"][0]
    assert payload["suite"] == "smoke"
    assert profile["generation_mode"] == "fixed_batched"
    assert "lineage_guardrails" not in profile
    assert set(profile["scenarios"]) == {
        "baseline",
        "throughput",
        "filtering",
        "missingness",
        "shift",
        "noise",
    }


def test_benchmark_cli_realizes_dataset_rows_once_per_preset(tmp_path) -> None:
    cfg = load_repo_config()
    cfg.dataset.rows = "400..60000"  # type: ignore[assignment]
    config_path = write_config(tmp_path, cfg, "rows_config.yaml")
    out_dir = tmp_path / "rows_benchmark"

    code = main(
        [
            "benchmark",
            "--config",
            str(config_path),
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "1",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--out-dir",
            str(out_dir),
        ]
    )

    assert code == 0
    payload = json.loads((out_dir / "summary.json").read_text(encoding="utf-8"))
    result = payload["preset_results"][0]
    effective_config = result["effective_config"]
    assert int(result["dataset_rows_total"]) <= int(SMOKE_N_TRAIN_CAP + SMOKE_N_TEST_CAP)
    assert effective_config["dataset"]["rows"] is None
    config_files = sorted(
        path
        for path in (out_dir / "effective_configs").glob("*.yaml")
        if not path.name.endswith("_trace.yaml")
    )
    assert config_files
    trace_files = sorted((out_dir / "effective_configs").glob("*_trace.yaml"))
    assert trace_files
    trace_payload = yaml.safe_load(trace_files[0].read_text(encoding="utf-8"))
    assert any(
        isinstance(item, dict) and item.get("source") == "benchmark.smoke_rows_cap"
        for item in trace_payload
    )


def test_effective_config_payload_yaml_matches_config_yaml() -> None:
    cfg = load_repo_config()

    assert effective_config_payload_yaml(effective_config_payload(cfg)) == effective_config_yaml(
        cfg
    )


def test_benchmark_cli_prints_effective_config_with_shared_yaml_shape(
    tmp_path,
    capsys: pytest.CaptureFixture[str],
) -> None:
    out_dir = tmp_path / "printed_effective_config"

    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "1",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--out-dir",
            str(out_dir),
            "--print-effective-config",
        ]
    )

    assert code == 0
    output = capsys.readouterr().out
    assert "Effective config [" in output
    assert "dataset:" in output
    assert "benchmark:" in output


@pytest.mark.parametrize(
    ("config_path", "scenario_name"),
    [
        ("configs/preset_filter_benchmark_smoke.yaml", "filtering"),
        ("configs/preset_missingness_mar.yaml", "missingness"),
        ("configs/preset_shift_benchmark_smoke.yaml", "shift"),
        ("configs/preset_noise_benchmark_smoke.yaml", "noise"),
    ],
)
def test_benchmark_cli_emits_expected_scenario(
    tmp_path, config_path: str, scenario_name: str
) -> None:
    out_dir = tmp_path / scenario_name
    code = main(
        [
            "benchmark",
            "--config",
            config_path,
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--out-dir",
            str(out_dir),
        ]
    )

    assert code == 0
    payload = json.loads((out_dir / "summary.json").read_text(encoding="utf-8"))
    scenario = payload["preset_results"][0]["scenarios"][scenario_name]
    assert scenario["enabled"] is True
    assert scenario["status"] in {"pass", "warn", "fail"}


def test_benchmark_cli_diagnostics_summary_omits_steering_payload(tmp_path) -> None:
    cfg = load_repo_config()
    cfg.diagnostics.enabled = True
    cfg.steering.enabled = True
    cfg.steering.preset = "anti_memorization_piecewise_v1"
    config_path = write_config(tmp_path, cfg, "diagnostics_config.yaml")
    out_dir = tmp_path / "bench_results"

    code = main(
        [
            "benchmark",
            "--config",
            str(config_path),
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--num-datasets",
            "1",
            "--warmup",
            "0",
            "--hardware-policy",
            "none",
            "--no-memory",
            "--diagnostics",
            "--out-dir",
            str(out_dir),
        ]
    )

    assert code == 0
    coverage_path = next((out_dir / "diagnostics").glob("*/coverage_summary.json"))
    coverage_summary = json.loads(coverage_path.read_text(encoding="utf-8"))
    assert "steering" not in coverage_summary
    assert "pearson_abs_mean" in coverage_summary["metrics"]


def test_print_preset_result_line_uses_scenario_statuses(
    capsys: pytest.CaptureFixture[str],
) -> None:
    _print_preset_result_line(
        {
            "preset_key": "cpu",
            "device": "cpu",
            "dataset_rows_total": 1024,
            "generation_mode": "fixed_batched",
            "hardware_backend": "cpu",
            "datasets_per_minute": 100.0,
            "generation_datasets_per_minute": 100.0,
            "write_datasets_per_minute": 80.0,
            "filter_datasets_per_minute": None,
            "filter_accepted_datasets_per_minute": None,
            "filter_rejection_rate_attempt_level": None,
            "filter_acceptance_rate_dataset_level": None,
            "filter_rejection_rate_dataset_level": None,
            "filter_retry_dataset_rate": None,
            "latency_p95_ms": 4.0,
            "scenarios": {
                "filtering": {"enabled": True, "status": "pass"},
                "missingness": {"enabled": True, "status": "fail"},
                "shift": {"enabled": False, "status": "off"},
                "noise": {"enabled": True, "status": "warn"},
                "throughput": {"enabled": True, "status": "pass"},
            },
        }
    )

    captured = capsys.readouterr()
    assert "filtering=pass" in captured.out
    assert "missingness=fail" in captured.out
    assert "noise=warn" in captured.out
    assert "throughput=pass" in captured.out



"""Component-level microbenchmarks for hot path sanity checks."""

from __future__ import annotations

import time
from collections.abc import Callable

import torch

from dagzoo.bench.constants import (
    MICRO_CONFIG_N_FEATURES_CAP,
    MICRO_CONFIG_N_NODES_CAP,
    MICRO_CONFIG_N_TEST_CAP,
    MICRO_CONFIG_N_TRAIN_CAP,
    MICROBENCH_DEFAULT_REPEATS,
    MICROBENCH_PARENT_DIM,
    MICROBENCH_SYNTH_FEATURES,
    MICROBENCH_SYNTH_ROWS,
    MILLISECONDS_PER_SECOND,
)
from dagzoo.config import GeneratorConfig, clone_generator_config, materialize_stress_profile
from dagzoo.core.dataset import generate_one
from dagzoo.core.fixed_layout.plan_types import FixedLayoutConverterSpec
from dagzoo.core.node_pipeline import apply_node_pipeline
from dagzoo.functions.random_functions import apply_random_function
from dagzoo.rng import KeyedRng

from .runtime_support import _synchronize_accelerator


def _time_ms(func: Callable[[], None], repeats: int, *, device: str | None = None) -> float:
    """Run ``func`` repeatedly and return the average runtime in milliseconds."""

    if device is not None:
        _synchronize_accelerator(device)
    start = time.perf_counter()
    for _ in range(max(1, repeats)):
        func()
    if device is not None:
        _synchronize_accelerator(device)
    elapsed = time.perf_counter() - start
    return (elapsed / max(1, repeats)) * MILLISECONDS_PER_SECOND


def _micro_config(config: GeneratorConfig) -> GeneratorConfig:
    """Create a lightweight config variant for inexpensive microbenchmarks."""

    if config.stress.profile is not None:
        config.validate_generation_constraints()
        c = materialize_stress_profile(
            config,
            revalidate=False,
            clear_selector=True,
        )
    else:
        c = clone_generator_config(config, revalidate=False)
    c.dataset.n_train = min(MICRO_CONFIG_N_TRAIN_CAP, c.dataset.n_train)
    c.dataset.n_test = min(MICRO_CONFIG_N_TEST_CAP, c.dataset.n_test)
    c.dataset.n_features_min = min(MICRO_CONFIG_N_FEATURES_CAP, c.dataset.n_features_min)
    c.dataset.n_features_max = min(MICRO_CONFIG_N_FEATURES_CAP, c.dataset.n_features_max)
    c.graph.n_nodes_min = min(MICRO_CONFIG_N_NODES_CAP, c.graph.n_nodes_min)
    c.graph.n_nodes_max = min(MICRO_CONFIG_N_NODES_CAP, c.graph.n_nodes_max)
    c.runtime.layout_mode = "fixed"
    return c


def run_microbenchmarks(
    config: GeneratorConfig,
    *,
    device: str | None = None,
    repeats: int = MICROBENCH_DEFAULT_REPEATS,
    include_generate_one: bool = True,
) -> dict[str, float | int | None]:
    """Run targeted microbenchmarks for core generation components."""

    micro_root = KeyedRng(int(config.seed)).keyed("bench", "micro")
    x = torch.randn(
        MICROBENCH_SYNTH_ROWS,
        MICROBENCH_SYNTH_FEATURES,
        generator=micro_root.keyed("inputs", "x").torch_rng(device="cpu"),
    )

    def run_linear() -> None:
        _ = apply_random_function(
            x,
            micro_root.keyed("random_function").torch_rng(device="cpu"),
            out_dim=MICROBENCH_SYNTH_FEATURES,
            function_type="linear",
        )

    parent_data = [
        torch.randn(
            MICROBENCH_SYNTH_ROWS,
            MICROBENCH_PARENT_DIM,
            generator=micro_root.keyed("inputs", "parent", 0).torch_rng(device="cpu"),
        ),
        torch.randn(
            MICROBENCH_SYNTH_ROWS,
            MICROBENCH_PARENT_DIM,
            generator=micro_root.keyed("inputs", "parent", 1).torch_rng(device="cpu"),
        ),
    ]
    specs = [
        FixedLayoutConverterSpec(
            key="feature_0", kind="num", dim=1, cardinality=None, column_start=0, column_end=1
        ),
        FixedLayoutConverterSpec(
            key="feature_1", kind="cat", dim=4, cardinality=6, column_start=1, column_end=5
        ),
        FixedLayoutConverterSpec(
            key="target", kind="target_cls", dim=3, cardinality=3, column_start=5, column_end=8
        ),
    ]

    def run_node_pipeline() -> None:
        _ = apply_node_pipeline(
            parent_data,
            MICROBENCH_SYNTH_ROWS,
            specs,
            micro_root.keyed("node_pipeline").torch_rng(device="cpu"),
            "cpu",
        )

    micro_cfg = _micro_config(config)
    generate_one_ms: float | None
    if include_generate_one:

        def run_generate_one() -> None:
            _ = generate_one(
                micro_cfg,
                seed=micro_root.child_seed("generate_one"),
                device=device,
            )

        generate_one_ms = _time_ms(
            run_generate_one, repeats, device=device or micro_cfg.runtime.device
        )
    else:
        generate_one_ms = None

    return {
        "micro_repeats": int(max(1, repeats)),
        "micro_random_function_linear_ms": _time_ms(run_linear, repeats),
        "micro_node_pipeline_ms": _time_ms(run_node_pipeline, repeats),
        "micro_generate_one_ms": generate_one_ms,
    }



"""Shared benchmark-style corpus probe helpers for audit workflows."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from dagzoo.bench.runtime_support import _resolve_benchmark_run_counts
from dagzoo.bench.stage_metrics import FilterStageMeasurement, replay_filter_stage_metrics
from dagzoo.bench.throughput import iter_throughput_measure_bundles, run_throughput_benchmark
from dagzoo.config import GeneratorConfig, clone_generator_config
from dagzoo.core.fixed_layout.runtime import realize_generation_config_for_run
from dagzoo.diagnostics.coverage import CoverageAggregationConfig, CoverageAggregator
from dagzoo.diagnostics_targets import build_diagnostics_aggregation_config

_REQUIRED_AUDIT_QUANTILES = (0.25, 0.5, 0.75)


@dataclass(slots=True)
class CorpusProbeResult:
    """Materialized one-config probe result for diversity audit workflows."""

    label: str
    config_path: str | None
    suite: str
    num_datasets: int
    warmup_datasets: int
    requested_device: str
    resolved_device: str
    resolved_config: dict[str, Any]
    datasets_per_minute: float
    filter_datasets_per_minute: float | None
    filter_accepted_datasets_per_minute: float | None
    filter_accepted_datasets_measured: int
    filter_rejected_datasets_measured: int
    filter_acceptance_rate_dataset_level: float | None
    filter_rejection_rate_dataset_level: float | None
    coverage_summary: dict[str, Any]
    filter_summary: dict[str, Any] | None


def resolve_corpus_probe_counts(
    config: GeneratorConfig,
    *,
    suite: str,
    num_datasets_override: int | None,
    warmup_override: int | None,
) -> tuple[int, int]:
    """Resolve one shared audit probe size using benchmark-style smoke caps."""

    if suite not in {"smoke", "standard"}:
        raise ValueError(f"Unsupported diversity-audit suite: {suite!r}")

    num_datasets = int(config.benchmark.num_datasets)
    warmup = int(config.benchmark.warmup_datasets)
    if num_datasets_override is not None:
        num_datasets = int(num_datasets_override)
    if warmup_override is not None:
        warmup = int(warmup_override)
    return _resolve_benchmark_run_counts(
        num_datasets=num_datasets,
        warmup_datasets=warmup,
        suite=suite,
    )


def build_corpus_probe_coverage_config(config: GeneratorConfig) -> CoverageAggregationConfig:
    """Build coverage aggregation config with quantiles needed for delta scoring."""

    base = build_diagnostics_aggregation_config(config)
    quantiles = tuple(
        sorted(set(float(q) for q in base.quantiles) | set(_REQUIRED_AUDIT_QUANTILES))
    )
    return CoverageAggregationConfig(
        include_spearman=bool(base.include_spearman),
        histogram_bins=max(1, int(base.histogram_bins)),
        quantiles=quantiles,
        underrepresented_threshold=float(base.underrepresented_threshold),
        max_values_per_metric=base.max_values_per_metric,
        target_bands=dict(base.target_bands),
    )


def _effective_probe_coverage_config(
    *,
    realized_config: GeneratorConfig,
    coverage_config: CoverageAggregationConfig | None,
) -> CoverageAggregationConfig:
    """Build one per-probe coverage config from shared settings plus realized steering."""

    template = (
        coverage_config
        if coverage_config is not None
        else build_corpus_probe_coverage_config(realized_config)
    )
    return CoverageAggregationConfig(
        include_spearman=bool(template.include_spearman),
        histogram_bins=max(1, int(template.histogram_bins)),
        quantiles=tuple(float(value) for value in template.quantiles),
        underrepresented_threshold=float(template.underrepresented_threshold),
        max_values_per_metric=template.max_values_per_metric,
        target_bands=dict(template.target_bands),
        steering_config=clone_generator_config(realized_config, revalidate=False),
    )


def _build_filter_summary(
    measurement: FilterStageMeasurement,
    *,
    config: GeneratorConfig,
) -> dict[str, Any]:
    """Normalize filter replay telemetry into the audit report shape."""

    return {
        "filter_mode": "small_shot_ease_v1",
        "ease_k_small": int(config.filter.ease_k_small),
        "easy_skill_threshold": float(config.filter.easy_skill_threshold),
        "easy_gain_threshold": float(config.filter.easy_gain_threshold),
        "hard_skill_threshold": float(config.filter.hard_skill_threshold),
        "stump_skill_threshold": (
            None
            if config.filter.stump_skill_threshold is None
            else float(config.filter.stump_skill_threshold)
        ),
        "use_lineage_veto": bool(config.filter.use_lineage_veto),
        "min_target_indegree": int(config.filter.min_target_indegree),
        "min_target_relevant_feature_count": int(config.filter.min_target_relevant_feature_count),
        "min_target_relevant_feature_fraction": float(
            config.filter.min_target_relevant_feature_fraction
        ),
        "classification_kappa_threshold": float(config.filter.classification_kappa_threshold),
        "classification_require_prediction_diversity": bool(
            config.filter.classification_require_prediction_diversity
        ),
        "accepted_true_fraction": measurement.accepted_true_fraction,
        "skill_small_mean": measurement.skill_small_mean,
        "skill_full_mean": measurement.skill_full_mean,
        "skill_gain_mean": measurement.skill_gain_mean,
        "skill_small_lb95_mean": measurement.skill_small_lb95_mean,
        "skill_gain_ub95_mean": measurement.skill_gain_ub95_mean,
        "skill_full_ub95_mean": measurement.skill_full_ub95_mean,
        "stump_skill_mean": measurement.stump_skill_mean,
        "reason_counts": dict(sorted(measurement.reason_counts.items())),
    }


def run_corpus_probe(
    config: GeneratorConfig,
    *,
    label: str,
    config_path: str | None,
    suite: str,
    num_datasets: int,
    warmup: int,
    device: str | None,
    probe_seed: int | None = None,
    coverage_config: CoverageAggregationConfig | None = None,
) -> CorpusProbeResult:
    """Run one benchmark-style corpus probe for diversity/comparison workflows."""

    realized_config, run_seed, requested_device, resolved_device = (
        realize_generation_config_for_run(
            config,
            seed=int(config.seed if probe_seed is None else probe_seed),
            device=device,
        )
    )
    realized_config.seed = int(run_seed)
    generation_config = clone_generator_config(realized_config, revalidate=False)
    generation_config.seed = int(run_seed)
    generation_config.filter.enabled = False

    throughput = run_throughput_benchmark(
        generation_config,
        num_datasets=num_datasets,
        warmup_datasets=warmup,
        device=requested_device,
        benchmark_fast_prepare=False,
    )

    coverage_aggregator = CoverageAggregator(
        _effective_probe_coverage_config(
            realized_config=realized_config,
            coverage_config=coverage_config,
        )
    )
    filter_datasets_per_minute: float | None = None
    filter_accepted_datasets_per_minute: float | None = None
    filter_accepted_datasets_measured = 0
    filter_rejected_datasets_measured = 0
    filter_acceptance_rate_dataset_level: float | None = None
    filter_rejection_rate_dataset_level: float | None = None
    filter_summary: dict[str, Any] | None = None
    analysis_stream = iter_throughput_measure_bundles(
        generation_config,
        num_datasets=num_datasets,
        device=requested_device,
        benchmark_fast_prepare=False,
    )

    if bool(realized_config.filter.enabled):

        def _record_accepted_bundle(bundle) -> None:
            coverage_aggregator.update_bundle(bundle)

        filter_measurement = replay_filter_stage_metrics(
            analysis_stream,
            config=realized_config,
            on_accepted_bundle=_record_accepted_bundle,
        )
        filter_datasets_per_minute = float(filter_measurement.datasets_per_minute)
        filter_accepted_datasets_measured = int(filter_measurement.filter_accepted_datasets)
        filter_rejected_datasets_measured = int(filter_measurement.filter_rejected_datasets)
        total_measured = filter_accepted_datasets_measured + filter_rejected_datasets_measured
        filter_acceptance_rate_dataset_level = (
            float(filter_accepted_datasets_measured) / float(total_measured)
            if total_measured > 0
            else None
        )
        filter_rejection_rate_dataset_level = (
            float(filter_rejected_datasets_measured) / float(total_measured)
            if total_measured > 0
            else None
        )
        filter_accepted_datasets_per_minute = (
            float(filter_datasets_per_minute) * float(filter_acceptance_rate_dataset_level)
            if filter_acceptance_rate_dataset_level is not None
            else None
        )
        filter_summary = _build_filter_summary(filter_measurement, config=realized_config)
    else:
        for bundle in analysis_stream:
            coverage_aggregator.update_bundle(bundle)

    coverage_summary = coverage_aggregator.build_summary()
    return CorpusProbeResult(
        label=label,
        config_path=config_path,
        suite=str(suite),
        num_datasets=int(num_datasets),
        warmup_datasets=int(warmup),
        requested_device=str(requested_device),
        resolved_device=str(resolved_device),
        resolved_config=realized_config.to_dict(),
        datasets_per_minute=float(throughput.get("datasets_per_minute", 0.0)),
        filter_datasets_per_minute=filter_datasets_per_minute,
        filter_accepted_datasets_per_minute=filter_accepted_datasets_per_minute,
        filter_accepted_datasets_measured=int(filter_accepted_datasets_measured),
        filter_rejected_datasets_measured=int(filter_rejected_datasets_measured),
        filter_acceptance_rate_dataset_level=filter_acceptance_rate_dataset_level,
        filter_rejection_rate_dataset_level=filter_rejection_rate_dataset_level,
        coverage_summary=coverage_summary,
        filter_summary=filter_summary,
    )



"""Multi-input function composition."""

from __future__ import annotations

import torch

from dagzoo.core.execution_semantics import sample_multi_source_plan
from dagzoo.core.fixed_layout.batched import FixedLayoutBatchRng, apply_function_plan_batch
from dagzoo.core.fixed_layout.plan_types import ConcatNodeSource, StackedNodeSource
from dagzoo.core.layout_types import AggregationKind, MechanismFamily
from dagzoo.functions.random_functions import apply_random_function
from dagzoo.math import sanitize_and_standardize
from dagzoo.rng import keyed_rng_from_generator
from dagzoo.sampling.noise import NoiseSamplingSpec


def _aggregate_parent_outputs(
    stacked: torch.Tensor,
    *,
    aggregation_kind: AggregationKind,
) -> torch.Tensor:
    """Aggregate transformed parent outputs using one explicit aggregation kind."""

    if aggregation_kind == "sum":
        return torch.sum(stacked, dim=1)
    if aggregation_kind == "product":
        return torch.prod(stacked, dim=1)
    if aggregation_kind == "max":
        return torch.max(stacked, dim=1).values
    if aggregation_kind == "logsumexp":
        return torch.logsumexp(stacked, dim=1)
    raise ValueError(f"Unknown aggregation kind: {aggregation_kind!r}")


def _aggregate_incrementally(
    aggregate: torch.Tensor,
    transformed_output: torch.Tensor,
    *,
    aggregation_kind: AggregationKind,
) -> torch.Tensor:
    """Update an aggregate tensor without materializing a full stack."""

    if aggregation_kind == "sum":
        return aggregate + transformed_output
    if aggregation_kind == "product":
        return aggregate * transformed_output
    if aggregation_kind == "max":
        return torch.maximum(aggregate, transformed_output)
    raise ValueError(f"Unknown aggregation kind: {aggregation_kind!r}")


def apply_multi_function(
    inputs: list[torch.Tensor],
    generator: torch.Generator,
    *,
    out_dim: int,
    aggregation_kind: AggregationKind | None = None,
    mechanism_logit_tilt: float = 0.0,
    function_family_mix: dict[MechanismFamily, float] | None = None,
    noise_sigma_multiplier: float = 1.0,
    noise_spec: NoiseSamplingSpec | None = None,
) -> torch.Tensor:
    """Apply composition across parent node tensors in torch.

    ``aggregation_kind`` only takes effect when the stacking composition path
    is selected (50 % chance).  When the concatenation path is taken, inputs
    are concatenated and passed through a single ``apply_random_function``,
    so the aggregation parameter is not used.
    """
    if not inputs:
        raise ValueError("inputs must be non-empty")
    if len(inputs) == 1:
        return apply_random_function(
            inputs[0],
            generator,
            out_dim=out_dim,
            mechanism_logit_tilt=mechanism_logit_tilt,
            function_family_mix=function_family_mix,
            noise_sigma_multiplier=noise_sigma_multiplier,
            noise_spec=noise_spec,
        )

    root = keyed_rng_from_generator(generator, "apply_multi_function")
    source = sample_multi_source_plan(
        keyed_rng=root.keyed("plan"),
        parent_count=len(inputs),
        out_dim=out_dim,
        mechanism_logit_tilt=mechanism_logit_tilt,
        function_family_mix=function_family_mix,
        aggregation_kind=aggregation_kind,
        device=str(generator.device),
    )
    rng = FixedLayoutBatchRng.from_keyed_rng(
        root.keyed("execution"),
        batch_size=1,
        device=str(inputs[0].device),
    )

    if isinstance(source, ConcatNodeSource):
        concat = sanitize_and_standardize(torch.cat(inputs, dim=1).to(torch.float32))
        out = apply_function_plan_batch(
            concat.unsqueeze(0),
            rng.keyed("function"),
            source.function,
            out_dim=out_dim,
            noise_sigma_multiplier=noise_sigma_multiplier,
            noise_spec=noise_spec,
            standardize_input=False,
        )
        return out.squeeze(0)

    if not isinstance(source, StackedNodeSource):
        raise RuntimeError("Expected a stacked multi-source plan.")

    if source.aggregation_kind == "logsumexp":
        transformed_outputs = [
            apply_function_plan_batch(
                sanitize_and_standardize(inp.to(torch.float32)).unsqueeze(0),
                rng.keyed("parent", plan_index),
                source.parent_functions[plan_index],
                out_dim=out_dim,
                noise_sigma_multiplier=noise_sigma_multiplier,
                noise_spec=noise_spec,
                standardize_input=False,
            ).squeeze(0)
            for plan_index, inp in enumerate(inputs)
        ]
        stacked = torch.stack(transformed_outputs, dim=1)  # (N, parents, out_dim)
        return _aggregate_parent_outputs(stacked, aggregation_kind=source.aggregation_kind)

    aggregate: torch.Tensor | None = None
    for plan_index, inp in enumerate(inputs):
        transformed_output = apply_function_plan_batch(
            sanitize_and_standardize(inp.to(torch.float32)).unsqueeze(0),
            rng.keyed("parent", plan_index),
            source.parent_functions[plan_index],
            out_dim=out_dim,
            noise_sigma_multiplier=noise_sigma_multiplier,
            noise_spec=noise_spec,
            standardize_input=False,
        ).squeeze(0)
        if aggregate is None:
            aggregate = transformed_output
        else:
            aggregate = _aggregate_incrementally(
                aggregate,
                transformed_output,
                aggregation_kind=source.aggregation_kind,
            )
    if aggregate is None:
        raise RuntimeError("Expected at least one parent transform for stacked multi-source.")
    return aggregate



"""Internal typed layout contracts for generation modules."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal

import torch

FeatureType = Literal["num", "cat"]
ConverterKind = Literal["cat", "num", "target_cls", "target_reg"]
MechanismFamily = Literal[
    "nn",
    "tree",
    "discretization",
    "gp",
    "linear",
    "quadratic",
    "em",
    "product",
    "piecewise",
]
AggregationKind = Literal["sum", "product", "max", "logsumexp"]


@dataclass(slots=True)
class LayoutPlan:
    """Typed sampled layout payload shared across generation modules."""

    n_features: int
    n_cat: int
    cat_idx: list[int]
    cardinalities: list[int]
    card_by_feature: dict[int, int]
    n_classes: int
    feature_types: list[FeatureType]
    graph_nodes: int
    graph_edges: int
    graph_depth_nodes: int
    graph_edge_density: float
    adjacency: torch.Tensor
    feature_node_assignment: list[int]
    target_to_node: int


__all__ = [
    "AggregationKind",
    "ConverterKind",
    "FeatureType",
    "LayoutPlan",
    "MechanismFamily",
]



from pathlib import Path
from types import SimpleNamespace

import pytest

from dagzoo.cli.entrypoint import main


def test_recipe_list_cli_prints_curated_catalog(capsys: pytest.CaptureFixture[str]) -> None:
    code = main(["recipe", "list"])

    assert code == 0
    captured = capsys.readouterr()
    assert "Curated dagzoo recipes" in captured.out
    assert "recipe:default-baseline" in captured.out
    assert "tabpfn-v1-prior-approx" in captured.out
    assert "category=reference prior" in captured.out


def test_generate_cli_prints_effective_config_and_resolution_trace(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    def _stub_generate_batch_iter(
        _config,
        *,
        num_datasets: int,
        seed: int | None = None,
        device: str | None = None,
    ):
        _ = seed
        _ = device
        for _ in range(num_datasets):
            yield object()

    monkeypatch.setattr(
        "dagzoo.cli.commands.generate.generate_batch_iter",
        _stub_generate_batch_iter,
    )

    code = main(
        [
            "generate",
            "--config",
            "configs/default.yaml",
            "--out",
            str(tmp_path / "generate"),
            "--num-datasets",
            "1",
            "--device",
            "cpu",
            "--hardware-policy",
            "none",
            "--no-dataset-write",
            "--print-effective-config",
            "--print-resolution-trace",
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Effective config:" in captured.out
    assert "Resolution trace:" in captured.out


def test_filter_cli_prints_curated_output_summary(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        "dagzoo.cli.commands.filter.run_deferred_filter",
        lambda **_kwargs: SimpleNamespace(
            manifest_path=tmp_path / "filter_out" / "filter_manifest.ndjson",
            summary_path=tmp_path / "filter_out" / "filter_summary.json",
            total_datasets=4,
            accepted_datasets=3,
            rejected_datasets=1,
            elapsed_seconds=2.0,
            datasets_per_minute=120.0,
            curated_out_dir=tmp_path / "curated_out",
            curated_accepted_datasets=3,
        ),
    )

    code = main(
        [
            "filter",
            "--in",
            "input_shards",
            "--out",
            str(tmp_path / "filter_out"),
            "--curated-out",
            str(tmp_path / "curated_out"),
            "--set",
            "filter.easy_skill_threshold=0.7",
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Wrote filter manifest:" in captured.out
    assert "Wrote filter summary:" in captured.out
    assert "Deferred filter summary: total=4 accepted=3 rejected=1 dpm=120.00" in captured.out
    assert "Wrote curated accepted-only shards:" in captured.out


def test_generate_cli_prints_handoff_execution_summary(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    def _stub_generate_batch_iter(*_args, **_kwargs):
        return iter(())

    def _stub_write_packed_parquet_shards_stream(
        _stream,
        *,
        out_dir: Path,
        shard_size: int,
        compression: str,
        internal_root: Path | None = None,
    ) -> int:
        assert shard_size > 0
        assert compression
        out_dir.mkdir(parents=True, exist_ok=True)
        if internal_root is not None:
            internal_root.mkdir(parents=True, exist_ok=True)
        return 2

    monkeypatch.setattr(
        "dagzoo.cli.commands.generate.generate_batch_iter",
        _stub_generate_batch_iter,
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.generate.write_packed_parquet_shards_stream",
        _stub_write_packed_parquet_shards_stream,
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.generate.write_generate_handoff_manifest",
        lambda **_kwargs: Path("handoff_manifest.json"),
    )

    code = main(
        [
            "generate",
            "--config",
            "configs/default.yaml",
            "--handoff-root",
            str(tmp_path / "handoff"),
            "--num-datasets",
            "2",
            "--device",
            "cpu",
            "--hardware-policy",
            "none",
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Wrote effective config:" in captured.out
    assert "Wrote effective config trace:" in captured.out
    assert "Wrote handoff manifest:" in captured.out
    assert "Wrote 2 datasets to:" in captured.out


def test_benchmark_cli_prints_configs_and_writes_baseline(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    summary = {
        "preset_results": [
            {
                "preset_key": "custom",
                "effective_config": {"runtime": {"device": "cpu"}},
                "effective_config_trace": [
                    {
                        "path": "runtime.device",
                        "source": "config",
                        "old_value": "auto",
                        "new_value": "cpu",
                    }
                ],
                "datasets_per_minute": 1.0,
                "latency_p95_ms": 1.0,
            }
        ],
        "regression": {"status": "pass", "issues": [], "hard_fail": False},
    }

    monkeypatch.setattr(
        "dagzoo.cli.commands.benchmark.run_benchmark_suite",
        lambda *args, **kwargs: summary,
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.benchmark.write_suite_json", lambda _summary, path: Path(path)
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.benchmark._print_preset_result_line",
        lambda _result: None,
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.benchmark.build_baseline_payload",
        lambda _summary: {"schema_version": 1},
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.benchmark.write_baseline",
        lambda _payload, path: Path(path),
    )

    code = main(
        [
            "benchmark",
            "--config",
            "configs/default.yaml",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--json-out",
            str(tmp_path / "summary.json"),
            "--save-baseline",
            str(tmp_path / "baseline.json"),
            "--print-effective-config",
            "--print-resolution-trace",
            "--no-memory",
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Effective config [custom]:" in captured.out
    assert "Resolution trace [custom]:" in captured.out
    assert "Wrote benchmark baseline:" in captured.out


def test_benchmark_cli_accepts_recipe_reference_for_custom_preset(
    tmp_path, monkeypatch: pytest.MonkeyPatch
) -> None:
    def _stub_run_benchmark_suite(preset_specs, *args, **kwargs):
        _ = args
        _ = kwargs
        assert len(preset_specs) == 1
        assert preset_specs[0].key == "default-baseline"
        return {
            "preset_results": [],
            "regression": {"status": "pass", "issues": [], "hard_fail": False},
        }

    monkeypatch.setattr(
        "dagzoo.cli.commands.benchmark.run_benchmark_suite",
        _stub_run_benchmark_suite,
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.benchmark.write_suite_json", lambda _summary, path: Path(path)
    )

    code = main(
        [
            "benchmark",
            "--config",
            "recipe:default-baseline",
            "--preset",
            "custom",
            "--suite",
            "smoke",
            "--json-out",
            str(tmp_path / "summary.json"),
            "--no-memory",
        ]
    )

    assert code == 0


def test_diversity_audit_cli_writes_summary_and_status(
    tmp_path, capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    report = {
        "summary": {"overall_status": "warn", "num_variants": 1},
    }

    monkeypatch.setattr(
        "dagzoo.cli.commands.diagnostics.run_effective_diversity_audit",
        lambda **kwargs: report,
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.diagnostics.write_effective_diversity_artifacts",
        lambda _report, out_dir: {"summary_json": Path(out_dir) / "summary.json"},
    )

    code = main(
        [
            "diversity-audit",
            "--baseline-config",
            "configs/default.yaml",
            "--variant-config",
            "configs/preset_shift_benchmark_smoke.yaml",
            "--out-dir",
            str(tmp_path / "diversity"),
        ]
    )

    assert code == 0
    captured = capsys.readouterr()
    assert "Wrote diversity artifact [summary_json]:" in captured.out
    assert "Diversity audit status=warn variants=1" in captured.out


def test_diversity_audit_cli_fail_on_regression_treats_insufficient_metrics_as_error(
    tmp_path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        "dagzoo.cli.commands.diagnostics.run_effective_diversity_audit",
        lambda **_kwargs: {
            "summary": {"overall_status": "insufficient_metrics", "num_variants": 1}
        },
    )
    monkeypatch.setattr(
        "dagzoo.cli.commands.diagnostics.write_effective_diversity_artifacts",
        lambda _report, out_dir: {"summary_json": Path(out_dir) / "summary.json"},
    )

    code = main(
        [
            "diversity-audit",
            "--baseline-config",
            "configs/default.yaml",
            "--variant-config",
            "configs/preset_shift_benchmark_smoke.yaml",
            "--fail-on-regression",
            "--out-dir",
            str(tmp_path / "diversity"),
        ]
    )

    assert code == 1


def test_hardware_cli_prints_detected_hardware(
    capsys: pytest.CaptureFixture[str], monkeypatch: pytest.MonkeyPatch
) -> None:
    class _Hardware:
        backend = "cpu"
        device_name = "cpu"
        tier = "cpu"
        total_memory_gb = 8.0
        peak_flops = 1.23e12

    monkeypatch.setattr("dagzoo.cli.commands.hardware.detect_hardware", lambda _device: _Hardware())

    code = main(["hardware", "--device", "cpu"])

    assert code == 0
    captured = capsys.readouterr()
    assert "backend=cpu" in captured.out
    assert "tier=cpu" in captured.out



"""Random positive weights sampling."""

import math

import torch

from dagzoo.math import log_uniform as _log_uniform
from dagzoo.sampling.noise import NoiseSamplingSpec, sample_noise_from_spec

_LOG_WEIGHT_CLAMP = 60.0


def sample_random_weights(
    dim: int,
    generator: torch.Generator,
    device: str,
    *,
    min_q_scale: float = 0.1,
    max_q: float = 6.0,
    sigma_min: float = 1e-4,
    sigma_max: float = 10.0,
    q: float | None = None,
    sigma: float | None = None,
    sigma_multiplier: float = 1.0,
    noise_spec: NoiseSamplingSpec | None = None,
) -> torch.Tensor:
    """Sample positive normalized weights using torch."""
    if dim <= 0:
        raise ValueError(f"dim must be > 0, got {dim}")
    if not math.isfinite(float(sigma_multiplier)) or float(sigma_multiplier) <= 0.0:
        raise ValueError(f"sigma_multiplier must be a finite value > 0, got {sigma_multiplier!r}")

    m = torch.arange(1, dim + 1, dtype=torch.float32, device=device)
    if q is None:
        q_low = min_q_scale / math.log(dim + 1.0)
        q = _log_uniform(generator, q_low, max_q, device)
    if sigma is None:
        sigma = _log_uniform(generator, sigma_min, sigma_max, device)
    effective_sigma = float(sigma) * float(sigma_multiplier)

    noise = sample_noise_from_spec(
        (dim,),
        generator=generator,
        device=device,
        noise_spec=noise_spec,
        scale_multiplier=effective_sigma,
    )
    log_w = (-float(q) * torch.log(m)) + noise
    log_w = torch.nan_to_num(
        log_w,
        nan=0.0,
        posinf=_LOG_WEIGHT_CLAMP,
        neginf=-_LOG_WEIGHT_CLAMP,
    )
    log_w = torch.clamp(log_w, min=-_LOG_WEIGHT_CLAMP, max=_LOG_WEIGHT_CLAMP)
    log_w = log_w - torch.max(log_w)
    w = torch.exp(log_w)
    w = torch.clamp(w, min=1e-12)
    w /= torch.sum(w)

    # Shuffle
    idx = torch.randperm(dim, generator=generator, device=device)
    return w[idx]



"""Shared generation-context helpers for seed, split, and device resolution."""

from __future__ import annotations

import torch

from dagzoo.config import (
    GeneratorConfig,
    resolve_dataset_total_rows,
    validate_class_split_feasibility,
)
from dagzoo.core.layout_types import LayoutPlan
from dagzoo.rng import validate_seed32


def _resolve_split_sizes(
    config: GeneratorConfig,
    *,
    dataset_seed: int | None = None,
) -> tuple[int, int]:
    """Resolve train/test split sizes from explicit or rows-derived config state."""

    n_test = int(config.dataset.n_test)
    total_rows = resolve_dataset_total_rows(config.dataset.rows, dataset_seed=dataset_seed)
    if total_rows is None:
        return int(config.dataset.n_train), n_test

    n_train = int(total_rows) - n_test
    if n_train <= 0:
        raise ValueError(
            "Resolved rows split is invalid: total rows must be > dataset.n_test "
            f"(total_rows={int(total_rows)}, n_test={n_test})."
        )
    return n_train, n_test


def _resolve_run_seed(config: GeneratorConfig, seed_override: int | None) -> int:
    """Resolve and validate the run seed used by generation entrypoints."""

    if seed_override is None:
        return validate_seed32(config.seed, field_name="seed")
    return validate_seed32(seed_override, field_name="seed")


def _resolve_device(config: GeneratorConfig, device_override: str | None) -> str:
    """Resolve runtime device and hard-fail on unavailable explicit accelerators."""

    requested = (device_override or config.runtime.device or "auto").lower()
    mps_ok = bool(getattr(torch.backends, "mps", None) and torch.backends.mps.is_available())
    if requested == "auto":
        if torch.cuda.is_available():
            return "cuda"
        if mps_ok:
            return "mps"
        return "cpu"
    if requested == "cpu":
        return "cpu"
    if requested == "cuda":
        if torch.cuda.is_available():
            return "cuda"
        raise RuntimeError("Requested device 'cuda' but CUDA is not available.")
    if requested == "mps":
        if mps_ok:
            return "mps"
        raise RuntimeError("Requested device 'mps' but MPS is not available.")
    raise ValueError(f"Unsupported device '{requested}'. Expected one of: auto, cpu, cuda, mps.")


def _torch_dtype(config: GeneratorConfig) -> torch.dtype:
    """Map string runtime dtype configuration to a torch dtype."""

    return torch.float64 if config.runtime.torch_dtype == "float64" else torch.float32


def _validate_class_split_for_layout(
    config: GeneratorConfig,
    *,
    layout: LayoutPlan,
    n_train: int,
    n_test: int,
) -> None:
    """Validate class/split feasibility for one sampled layout/split pair."""

    if config.dataset.task != "classification":
        return
    validate_class_split_feasibility(
        n_classes=int(layout.n_classes),
        n_train=int(n_train),
        n_test=int(n_test),
        context=(
            "sampled classification split constraints "
            f"(n_train={int(n_train)}, n_test={int(n_test)})"
        ),
    )



"""Run-level coverage aggregation and artifact writers."""

from __future__ import annotations

import datetime as dt
import json
import math
import random
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

import numpy as np

from dagzoo.config import GeneratorConfig
from dagzoo.math import sanitize_json as _sanitize_json
from dagzoo.types import DatasetBundle

from .metrics import extract_dataset_metrics
from .types import DatasetMetrics

_DEFAULT_QUANTILES = (0.05, 0.25, 0.5, 0.75, 0.95)
_DEFAULT_MAX_VALUES_PER_METRIC = 50_000
_STEERING_MIXTURE_COMPONENTS = ("gaussian", "laplace", "student_t")
_METRIC_FIELD_NAMES = (
    "n_rows",
    "n_features",
    "n_classes",
    "categorical_ratio",
    "pearson_abs_mean",
    "pearson_abs_max",
    "class_entropy",
    "majority_minority_ratio",
    "snr_proxy_db",
    "cat_cardinality_mean",
)


@dataclass(slots=True)
class CoverageAggregationConfig:
    """Configuration for run-level coverage aggregation."""

    include_spearman: bool = False
    histogram_bins: int = 10
    quantiles: tuple[float, ...] = _DEFAULT_QUANTILES
    underrepresented_threshold: float = 0.5
    max_values_per_metric: int | None = _DEFAULT_MAX_VALUES_PER_METRIC
    target_bands: dict[str, tuple[float, float]] = field(default_factory=dict)
    steering_config: GeneratorConfig | None = None


@dataclass(slots=True)
class _ValueAccumulator:
    count: int = 0
    total: float = 0.0
    min_value: float = math.inf
    max_value: float = -math.inf

    def update(self, value: object) -> None:
        if isinstance(value, bool) or not isinstance(value, (int, float)):
            return
        as_float = float(value)
        if not math.isfinite(as_float):
            return
        self.count += 1
        self.total += as_float
        self.min_value = min(self.min_value, as_float)
        self.max_value = max(self.max_value, as_float)

    def finalize(self) -> dict[str, Any]:
        if self.count <= 0:
            return {"count": 0, "min": None, "max": None, "mean": None}
        return {
            "count": int(self.count),
            "min": float(self.min_value),
            "max": float(self.max_value),
            "mean": float(self.total / self.count),
        }


@dataclass(slots=True)
class _MetricAccumulator:
    count: int = 0
    missing_count: int = 0
    total: float = 0.0
    total_sq: float = 0.0
    min_value: float = math.inf
    max_value: float = -math.inf
    values: list[float] = field(default_factory=list)
    sample_limit: int | None = None
    rng_seed: int = 0
    _seen_values: int = 0
    _rng: random.Random = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self._rng = random.Random(self.rng_seed)

    def update(self, value: float | int | None) -> None:
        if value is None:
            self.missing_count += 1
            return
        as_float = float(value)
        if not math.isfinite(as_float):
            self.missing_count += 1
            return
        self.count += 1
        self.total += as_float
        self.total_sq += as_float * as_float
        self.min_value = min(self.min_value, as_float)
        self.max_value = max(self.max_value, as_float)
        self._seen_values += 1
        if self.sample_limit is None:
            self.values.append(as_float)
            return
        if len(self.values) < self.sample_limit:
            self.values.append(as_float)
            return
        # Deterministic reservoir sampling bounds memory while preserving an unbiased sample.
        replace_idx = self._rng.randint(0, self._seen_values - 1)
        if replace_idx < self.sample_limit:
            self.values[replace_idx] = as_float

    def finalize(
        self,
        *,
        quantiles: tuple[float, ...],
        histogram_bins: int,
        underrepresented_threshold: float,
        target_band: tuple[float, float] | None,
    ) -> dict[str, Any]:
        if self.count <= 0:
            return {
                "count": 0,
                "missing_count": int(self.missing_count),
                "observed_min": None,
                "observed_max": None,
                "mean": None,
                "std": None,
                "sampled_count": 0,
                "sampled_fraction": 0.0,
                "quantiles": {f"p{int(round(q * 100)):02d}": None for q in quantiles},
                "histogram": {
                    "num_bins": int(histogram_bins),
                    "covered_bins": 0,
                    "coverage_ratio": 0.0,
                    "bins": [],
                },
                "underrepresented_bins": [],
                "target_band": _target_band_payload(target_band),
            }

        values = np.asarray(self.values, dtype=np.float64)
        mean = float(self.total / self.count)
        variance = max(0.0, float((self.total_sq / self.count) - (mean * mean)))
        std = math.sqrt(variance)
        sampled_count = int(values.size)
        quantile_map = {
            f"p{int(round(q * 100)):02d}": float(np.quantile(values, q)) for q in quantiles
        }

        histogram = _build_histogram(
            values,
            bins=histogram_bins,
            value_range=target_band,
        )
        underrepresented_bins: list[dict[str, Any]] = []
        target_payload = _target_band_payload(target_band)
        if target_band is not None:
            underrepresented_bins, in_target_count, in_target_fraction = _underrepresented_bins(
                values=values,
                bins=histogram["bins"],
                target_band=target_band,
                underrepresented_threshold=underrepresented_threshold,
            )
            if target_payload is not None:
                target_payload["in_target_count"] = in_target_count
                target_payload["in_target_fraction"] = in_target_fraction

        return {
            "count": int(self.count),
            "missing_count": int(self.missing_count),
            "observed_min": float(self.min_value),
            "observed_max": float(self.max_value),
            "mean": mean,
            "std": float(std),
            "sampled_count": sampled_count,
            "sampled_fraction": float(sampled_count / self.count) if self.count > 0 else 0.0,
            "quantiles": quantile_map,
            "histogram": histogram,
            "underrepresented_bins": underrepresented_bins,
            "target_band": target_payload,
        }


class CoverageAggregator:
    """Streaming aggregator for run-level dataset diagnostics coverage."""

    def __init__(self, config: CoverageAggregationConfig | None = None) -> None:
        cfg = config or CoverageAggregationConfig()
        self._config = CoverageAggregationConfig(
            include_spearman=bool(cfg.include_spearman),
            histogram_bins=max(1, int(cfg.histogram_bins)),
            quantiles=_normalize_quantiles(cfg.quantiles),
            underrepresented_threshold=max(0.0, float(cfg.underrepresented_threshold)),
            max_values_per_metric=_normalize_max_values_per_metric(cfg.max_values_per_metric),
            target_bands=_normalize_target_bands(cfg.target_bands),
            steering_config=cfg.steering_config,
        )
        self._num_datasets = 0
        self._task_counts: dict[str, int] = {}
        self._mechanism_family_bundles_with_metadata = 0
        self._mechanism_family_total_function_plans = 0
        self._mechanism_family_sampled_counts: dict[str, int] = {}
        self._mechanism_family_dataset_presence: dict[str, int] = {}
        self._mechanism_variant_sampled_counts: dict[str, int] = {}
        self._mechanism_variant_dataset_presence: dict[str, int] = {}
        self._metrics = {
            name: _MetricAccumulator(
                sample_limit=self._config.max_values_per_metric,
                rng_seed=idx + 1,
            )
            for idx, name in enumerate(_METRIC_FIELD_NAMES)
        }

    @property
    def num_datasets(self) -> int:
        """Return number of ingested datasets."""

        return int(self._num_datasets)

    def update_bundle(self, bundle: DatasetBundle) -> DatasetMetrics:
        """Extract metrics from one bundle and update aggregator state."""

        metrics = extract_dataset_metrics(bundle, include_spearman=self._config.include_spearman)
        self.update_metrics(metrics)
        self._update_mechanism_families(bundle)
        return metrics

    def update_metrics(self, metrics: DatasetMetrics) -> None:
        """Update aggregator state from one metrics payload."""

        self._num_datasets += 1
        self._task_counts[metrics.task] = self._task_counts.get(metrics.task, 0) + 1
        for metric_name in _METRIC_FIELD_NAMES:
            value = getattr(metrics, metric_name)
            self._metrics[metric_name].update(value)

    def build_summary(self) -> dict[str, Any]:
        """Build finalized run-level coverage summary."""

        summary_metrics: dict[str, Any] = {}
        for metric_name, accumulator in self._metrics.items():
            summary_metrics[metric_name] = accumulator.finalize(
                quantiles=self._config.quantiles,
                histogram_bins=self._config.histogram_bins,
                underrepresented_threshold=self._config.underrepresented_threshold,
                target_band=self._config.target_bands.get(metric_name),
            )

        return {
            "generated_at": dt.datetime.now(dt.UTC).isoformat(),
            "num_datasets": int(self._num_datasets),
            "task_counts": dict(sorted(self._task_counts.items())),
            "histogram_bins": int(self._config.histogram_bins),
            "quantiles": list(self._config.quantiles),
            "max_values_per_metric": self._config.max_values_per_metric,
            "mechanism_family_summary": self._build_mechanism_family_summary(),
            "metrics": summary_metrics,
        }

    def _update_mechanism_families(self, bundle: DatasetBundle) -> None:
        metadata = bundle.metadata.get("mechanism_families")
        if not isinstance(metadata, dict):
            return
        sampled_family_counts = metadata.get("sampled_family_counts")
        if not isinstance(sampled_family_counts, dict):
            return

        normalized_counts: dict[str, int] = {}
        for raw_family, raw_count in sampled_family_counts.items():
            if isinstance(raw_family, str) and not isinstance(raw_count, bool):
                if isinstance(raw_count, (int, float)) and math.isfinite(float(raw_count)):
                    count = max(0, int(raw_count))
                    if count > 0:
                        normalized_counts[str(raw_family)] = count
        sampled_variant_counts = metadata.get("sampled_variant_counts")
        normalized_variant_counts: dict[str, int] = {}
        if isinstance(sampled_variant_counts, dict):
            for raw_label, raw_count in sampled_variant_counts.items():
                if isinstance(raw_label, str) and not isinstance(raw_count, bool):
                    if isinstance(raw_count, (int, float)) and math.isfinite(float(raw_count)):
                        count = max(0, int(raw_count))
                        if count > 0:
                            normalized_variant_counts[str(raw_label)] = count
        self._mechanism_family_bundles_with_metadata += 1
        for family, count in normalized_counts.items():
            self._mechanism_family_sampled_counts[family] = (
                self._mechanism_family_sampled_counts.get(family, 0) + int(count)
            )
            self._mechanism_family_dataset_presence[family] = (
                self._mechanism_family_dataset_presence.get(family, 0) + 1
            )
        for label, count in normalized_variant_counts.items():
            self._mechanism_variant_sampled_counts[label] = (
                self._mechanism_variant_sampled_counts.get(label, 0) + int(count)
            )
            self._mechanism_variant_dataset_presence[label] = (
                self._mechanism_variant_dataset_presence.get(label, 0) + 1
            )

        total_function_plans = metadata.get("total_function_plans")
        if (
            not isinstance(total_function_plans, bool)
            and isinstance(total_function_plans, (int, float))
            and math.isfinite(float(total_function_plans))
        ):
            self._mechanism_family_total_function_plans += max(0, int(total_function_plans))

    def _build_mechanism_family_summary(self) -> dict[str, Any]:
        bundles_with_metadata = int(self._mechanism_family_bundles_with_metadata)
        num_datasets = int(self._num_datasets)
        metadata_denominator = num_datasets if num_datasets > 0 else 0
        mean_total_function_plans = (
            float(self._mechanism_family_total_function_plans / bundles_with_metadata)
            if bundles_with_metadata > 0
            else 0.0
        )
        return {
            "metadata_coverage_rate": (
                float(bundles_with_metadata / metadata_denominator)
                if metadata_denominator > 0
                else 0.0
            ),
            "bundles_with_metadata": bundles_with_metadata,
            "sampled_family_counts": dict(sorted(self._mechanism_family_sampled_counts.items())),
            "dataset_presence_rate_by_family": {
                family: (float(count / metadata_denominator) if metadata_denominator > 0 else 0.0)
                for family, count in sorted(self._mechanism_family_dataset_presence.items())
            },
            "sampled_variant_counts": dict(sorted(self._mechanism_variant_sampled_counts.items())),
            "dataset_presence_rate_by_variant": {
                label: (float(count / metadata_denominator) if metadata_denominator > 0 else 0.0)
                for label, count in sorted(self._mechanism_variant_dataset_presence.items())
            },
            "mean_total_function_plans": float(mean_total_function_plans),
        }


def write_coverage_summary_json(summary: dict[str, Any], out_path: str | Path) -> Path:
    """Write run-level coverage summary JSON artifact."""

    path = Path(out_path)
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open("w", encoding="utf-8") as f:
        json.dump(_sanitize_json(summary), f, indent=2, sort_keys=True, allow_nan=False)
    return path


def write_coverage_summary_markdown(summary: dict[str, Any], out_path: str | Path) -> Path:
    """Write a concise markdown artifact for run-level coverage."""

    path = Path(out_path)
    path.parent.mkdir(parents=True, exist_ok=True)

    lines = [
        "# Coverage Summary",
        "",
        f"- Generated at: `{summary.get('generated_at', '-')}`",
        f"- Num datasets: `{summary.get('num_datasets', 0)}`",
        f"- Histogram bins: `{summary.get('histogram_bins', 0)}`",
        f"- Quantiles: `{summary.get('quantiles', [])}`",
        f"- Max sampled values/metric: `{summary.get('max_values_per_metric', '-')}`",
    ]
    task_counts = summary.get("task_counts", {})
    if isinstance(task_counts, dict) and task_counts:
        task_parts = [f"{name}={count}" for name, count in sorted(task_counts.items())]
        lines.append(f"- Task counts: `{', '.join(task_parts)}`")
    mechanism_family_summary = summary.get("mechanism_family_summary", {})
    if isinstance(mechanism_family_summary, dict):
        lines.extend(
            [
                f"- Mechanism metadata coverage: `{_fmt(mechanism_family_summary.get('metadata_coverage_rate'))}`",
                f"- Bundles with mechanism metadata: `{_fmt(mechanism_family_summary.get('bundles_with_metadata'), digits=0)}`",
                f"- Mean total function plans: `{_fmt(mechanism_family_summary.get('mean_total_function_plans'))}`",
            ]
        )
    lines.extend(["", "## Metrics", ""])
    lines.append("| Metric | Min | Max | p50 | Covered Bins | Underrepresented Bins |")
    lines.append("|---|---:|---:|---:|---:|---:|")

    metrics_payload = summary.get("metrics", {})
    if isinstance(metrics_payload, dict):
        for metric_name in sorted(metrics_payload):
            metric = metrics_payload[metric_name]
            if not isinstance(metric, dict):
                continue
            histogram = metric.get("histogram", {})
            under_bins = metric.get("underrepresented_bins", [])
            lines.append(
                "| "
                f"{metric_name} | "
                f"{_fmt(metric.get('observed_min'))} | "
                f"{_fmt(metric.get('observed_max'))} | "
                f"{_fmt((metric.get('quantiles') or {}).get('p50'))} | "
                f"{_fmt((histogram or {}).get('covered_bins'), digits=0)} | "
                f"{_fmt(len(under_bins), digits=0)} |"
            )

    if isinstance(mechanism_family_summary, dict):
        lines.extend(["", "## Mechanism Families", ""])
        sampled_family_counts = mechanism_family_summary.get("sampled_family_counts", {})
        dataset_presence = mechanism_family_summary.get("dataset_presence_rate_by_family", {})
        if isinstance(sampled_family_counts, dict) and sampled_family_counts:
            lines.append("| Family | Sampled Count | Dataset Presence Rate |")
            lines.append("|---|---:|---:|")
            for family in sorted(sampled_family_counts):
                lines.append(
                    "| "
                    f"{family} | "
                    f"{_fmt(sampled_family_counts.get(family), digits=0)} | "
                    f"{_fmt((dataset_presence or {}).get(family))} |"
                )
        else:
            lines.append("- No realized mechanism-family metadata was observed.")
        sampled_variant_counts = mechanism_family_summary.get("sampled_variant_counts", {})
        variant_presence = mechanism_family_summary.get("dataset_presence_rate_by_variant", {})
        if isinstance(sampled_variant_counts, dict) and sampled_variant_counts:
            lines.extend(["", "### Mechanism Variants", ""])
            lines.append("| Variant | Sampled Count | Dataset Presence Rate |")
            lines.append("|---|---:|---:|")
            for label in sorted(sampled_variant_counts):
                lines.append(
                    "| "
                    f"{label} | "
                    f"{_fmt(sampled_variant_counts.get(label), digits=0)} | "
                    f"{_fmt((variant_presence or {}).get(label))} |"
                )

    with path.open("w", encoding="utf-8") as f:
        f.write("\n".join(lines).rstrip() + "\n")
    return path


def _normalize_quantiles(quantiles: tuple[float, ...] | list[float]) -> tuple[float, ...]:
    raw = list(quantiles) if quantiles else list(_DEFAULT_QUANTILES)
    normalized: list[float] = []
    for q in raw:
        value = float(q)
        if 0.0 <= value <= 1.0:
            normalized.append(value)
    if not normalized:
        return _DEFAULT_QUANTILES
    return tuple(sorted(set(normalized)))


def _normalize_max_values_per_metric(raw: object) -> int | None:
    if raw is None:
        return None
    if isinstance(raw, bool):
        return _DEFAULT_MAX_VALUES_PER_METRIC
    if isinstance(raw, int):
        value = raw
    elif isinstance(raw, float):
        if not math.isfinite(raw):
            return _DEFAULT_MAX_VALUES_PER_METRIC
        value = int(raw)
    elif isinstance(raw, str):
        try:
            value = int(raw)
        except ValueError:
            return _DEFAULT_MAX_VALUES_PER_METRIC
    else:
        return _DEFAULT_MAX_VALUES_PER_METRIC
    if value <= 0:
        return None
    return value


def _normalize_target_bands(
    target_bands: dict[str, Any],
) -> dict[str, tuple[float, float]]:
    normalized: dict[str, tuple[float, float]] = {}
    for metric_name, band in target_bands.items():
        if not isinstance(metric_name, str):
            continue
        if not isinstance(band, (list, tuple)) or len(band) != 2:
            continue
        lo = float(band[0])
        hi = float(band[1])
        if not math.isfinite(lo) or not math.isfinite(hi):
            continue
        if lo <= hi:
            normalized[metric_name] = (lo, hi)
        else:
            normalized[metric_name] = (hi, lo)
    return normalized


def _coerce_optional_int(value: object) -> int | None:
    if isinstance(value, bool) or not isinstance(value, (int, float)):
        return None
    as_float = float(value)
    if not math.isfinite(as_float):
        return None
    return int(as_float)


def _bundle_config_missing_rate(bundle: DatasetBundle) -> float | None:
    config_payload = bundle.metadata.get("config")
    if not isinstance(config_payload, dict):
        return None
    dataset_payload = config_payload.get("dataset")
    if not isinstance(dataset_payload, dict):
        return None
    return _coerce_optional_float(dataset_payload.get("missing_rate"))


def _coerce_optional_float(value: object) -> float | None:
    if isinstance(value, bool) or not isinstance(value, (int, float)):
        return None
    as_float = float(value)
    if not math.isfinite(as_float):
        return None
    return as_float


def _float_matches(lhs: object, rhs: object, *, atol: float = 1e-9, rtol: float = 1e-9) -> bool:
    lhs_float = _coerce_optional_float(lhs)
    rhs_float = _coerce_optional_float(rhs)
    if lhs_float is None or rhs_float is None:
        return lhs_float is None and rhs_float is None
    return math.isclose(lhs_float, rhs_float, abs_tol=atol, rel_tol=rtol)


def _fmt_counts(payload: object) -> str:
    if not isinstance(payload, dict) or not payload:
        return "-"
    parts = []
    for key in sorted(payload):
        value = payload.get(key)
        if isinstance(value, (int, float)) and not isinstance(value, bool):
            parts.append(f"{key}={int(value)}")
    return ", ".join(parts) if parts else "-"


def _fmt_int_range(payload: object) -> str:
    if not isinstance(payload, dict):
        return "-"
    lo = _coerce_optional_int(payload.get("min"))
    hi = _coerce_optional_int(payload.get("max"))
    if lo is None or hi is None:
        return "-"
    return f"{lo}-{hi}"


def _fmt_value_range(payload: object) -> str:
    if not isinstance(payload, dict):
        return "-"
    lo = _coerce_optional_float(payload.get("min"))
    hi = _coerce_optional_float(payload.get("max"))
    if lo is None or hi is None:
        return "-"
    return f"{lo:.3f}-{hi:.3f}"


def _fmt_scalar_summary(payload: object) -> str:
    if not isinstance(payload, dict):
        return "-"
    mean = _coerce_optional_float(payload.get("mean"))
    lo = _coerce_optional_float(payload.get("min"))
    hi = _coerce_optional_float(payload.get("max"))
    if mean is None or lo is None or hi is None:
        return "-"
    return f"{mean:.3f} ({lo:.3f}-{hi:.3f})"


def _build_histogram(
    values: np.ndarray,
    *,
    bins: int,
    value_range: tuple[float, float] | None = None,
) -> dict[str, Any]:
    if values.size <= 0:
        return {
            "num_bins": int(bins),
            "covered_bins": 0,
            "coverage_ratio": 0.0,
            "bins": [],
        }
    if value_range is None:
        v_min = float(np.min(values))
        v_max = float(np.max(values))
        if v_min == v_max:
            lo = v_min - 0.5
            hi = v_max + 0.5
        else:
            lo = v_min
            hi = v_max
    else:
        lo = float(value_range[0])
        hi = float(value_range[1])
        if lo > hi:
            lo, hi = hi, lo
    if not np.isfinite(lo) or not np.isfinite(hi):
        lo = -0.5
        hi = 0.5
    elif not hi > lo or np.isclose(lo, hi, rtol=1e-12, atol=1e-12):
        center = float((lo + hi) / 2.0)
        span = max(0.5, abs(center) * 1e-6)
        lo = center - span
        hi = center + span
    counts, edges = np.histogram(values, bins=bins, range=(lo, hi))
    total = float(np.sum(counts))
    bins_payload: list[dict[str, Any]] = []
    for i in range(bins):
        count = int(counts[i])
        bins_payload.append(
            {
                "index": int(i),
                "lower": float(edges[i]),
                "upper": float(edges[i + 1]),
                "count": count,
                "fraction": float(count / total) if total > 0 else 0.0,
            }
        )
    covered_bins = int(sum(1 for b in bins_payload if b["count"] > 0))
    return {
        "num_bins": int(bins),
        "covered_bins": covered_bins,
        "coverage_ratio": float(covered_bins / bins) if bins > 0 else 0.0,
        "bins": bins_payload,
    }


def _underrepresented_bins(
    values: np.ndarray,
    bins: list[dict[str, Any]],
    *,
    target_band: tuple[float, float],
    underrepresented_threshold: float,
) -> tuple[list[dict[str, Any]], int, float]:
    lo, hi = target_band
    in_target_mask = (values >= lo) & (values <= hi)
    in_target_count = int(np.sum(in_target_mask))
    total_count = int(values.size)
    in_target_fraction = float(in_target_count / total_count) if total_count > 0 else 0.0

    overlapping = [
        b
        for b in bins
        if isinstance(b.get("lower"), (int, float))
        and isinstance(b.get("upper"), (int, float))
        and not (float(b["upper"]) <= lo or float(b["lower"]) >= hi)
    ]
    if not overlapping:
        return [], in_target_count, in_target_fraction
    if in_target_count == 0:
        return overlapping, in_target_count, in_target_fraction

    edges = [float(bins[0]["lower"])] + [float(b["upper"]) for b in bins]
    in_target_values = values[in_target_mask]
    in_target_bin_counts, _ = np.histogram(
        in_target_values, bins=np.asarray(edges, dtype=np.float64)
    )
    expected_count_per_bin = float(in_target_count / len(overlapping))
    threshold_count = expected_count_per_bin * underrepresented_threshold

    underrepresented: list[dict[str, Any]] = []
    for b in overlapping:
        idx = int(b.get("index", -1))
        if idx < 0 or idx >= int(in_target_bin_counts.size):
            continue
        in_target_bin_count = int(in_target_bin_counts[idx])
        if float(in_target_bin_count) < threshold_count:
            annotated = dict(b)
            annotated["in_target_count"] = in_target_bin_count
            annotated["in_target_fraction"] = (
                float(in_target_bin_count / in_target_count) if in_target_count > 0 else 0.0
            )
            underrepresented.append(annotated)
    return underrepresented, in_target_count, in_target_fraction


def _target_band_payload(target_band: tuple[float, float] | None) -> dict[str, float] | None:
    if target_band is None:
        return None
    return {"min": float(target_band[0]), "max": float(target_band[1])}


def _fmt(value: Any, digits: int = 3) -> str:
    if not isinstance(value, (int, float)):
        return "-"
    as_float = float(value)
    if not math.isfinite(as_float):
        return "-"
    return f"{as_float:.{digits}f}"



"""Recipe catalog command handlers."""

from __future__ import annotations

import argparse

from dagzoo.recipes import iter_recipe_specs


def run_recipe_list_command(_args: argparse.Namespace) -> int:
    """Print the curated public recipe catalog."""

    print("Curated dagzoo recipes")
    print("Use with: dagzoo generate --config recipe:<name> --num-datasets <n> --out <dir>")
    for spec in iter_recipe_specs():
        print(
            f"- {spec.reference}: {spec.summary} "
            f"[category={spec.category}; confidence={spec.confidence_tier}; "
            f"regime={spec.expected_regime}]"
        )
    return 0



from __future__ import annotations

from dataclasses import dataclass

from .common import normalize_files, run_git_capture, run_git_lines
from .contract import ContractResult, evaluate_release_contract
from .impact import ImpactReport, build_impact_report

DEFAULT_BASE_REF = "origin/main"


@dataclass(frozen=True)
class ReviewScope:
    base_ref: str
    merge_base: str
    changed_files: tuple[str, ...]


@dataclass(frozen=True)
class ReviewBaseResult:
    scope: ReviewScope
    report: ImpactReport
    contract: ContractResult


def read_merge_base(base_ref: str = DEFAULT_BASE_REF) -> str:
    return run_git_capture("merge-base", base_ref, "HEAD").strip()


def collect_review_scope(base_ref: str = DEFAULT_BASE_REF) -> ReviewScope:
    merge_base = read_merge_base(base_ref)
    changed_files = tuple(
        sorted(
            dict.fromkeys(
                normalize_files(
                    (
                        *run_git_lines("diff", "--name-only", f"{merge_base}..HEAD"),
                        *run_git_lines("diff", "--name-only", "--cached"),
                        *run_git_lines("diff", "--name-only"),
                        *run_git_lines("ls-files", "--others", "--exclude-standard"),
                    )
                )
            )
        )
    )
    return ReviewScope(base_ref=base_ref, merge_base=merge_base, changed_files=changed_files)


def build_review_base_result(base_ref: str = DEFAULT_BASE_REF) -> ReviewBaseResult:
    scope = collect_review_scope(base_ref)
    report = build_impact_report(scope.changed_files)
    contract = evaluate_release_contract(report)
    return ReviewBaseResult(scope=scope, report=report, contract=contract)


def render_review_base_report(result: ReviewBaseResult) -> str:
    lines = [
        "review-base",
        f"base ref: {result.scope.base_ref}",
        f"merge base: {result.scope.merge_base}",
        "changed files:",
    ]
    if result.scope.changed_files:
        lines.extend(f"- `{path}`" for path in result.scope.changed_files)
    else:
        lines.append("- none")

    lines.extend(
        [
            "",
            "tags:",
            "- "
            + (
                ", ".join(f"`{tag}`" for tag in result.report.tags)
                if result.report.tags
                else "none"
            ),
            "",
            "recommended verify modes:",
            "- "
            + (
                ", ".join(f"`{mode}`" for mode in result.report.recommended_modes)
                if result.report.recommended_modes
                else "none"
            ),
        ]
    )
    if result.report.suggested_pytest_targets:
        lines.extend(
            [
                "",
                "suggested pytest targets:",
                "- "
                + ", ".join(f"`{target}`" for target in result.report.suggested_pytest_targets),
            ]
        )
    lines.extend(
        [
            "",
            "pytest selection:",
            f"- mode: `{result.report.pytest_selection.mode}`",
            f"- reason: {result.report.pytest_selection.reason}",
        ]
    )
    if result.report.pytest_selection.targets:
        lines.append(
            "- targets: "
            + ", ".join(f"`{target}`" for target in result.report.pytest_selection.targets)
        )

    lines.extend(["", "release contract:"])
    if result.contract.warnings:
        lines.extend(f"- warning: {warning}" for warning in result.contract.warnings)
    if result.contract.errors:
        lines.extend(f"- error: {error}" for error in result.contract.errors)
    if not result.contract.warnings and not result.contract.errors:
        lines.append("- release contract checks passed.")

    return "\n".join(lines) + "\n"



"""Shared math and random-matrix utilities."""

from .utils import (
    coerce_optional_finite_float,
    log_uniform,
    normalize_positive_weights,
    row_normalize,
    sanitize_and_standardize,
    sanitize_json,
    standardize,
    to_numpy,
)

__all__ = [
    "coerce_optional_finite_float",
    "log_uniform",
    "normalize_positive_weights",
    "row_normalize",
    "sanitize_and_standardize",
    "sanitize_json",
    "standardize",
    "to_numpy",
]



"""Shift runtime parameter resolution for sampling-path integration."""

from __future__ import annotations

import math
from dataclasses import dataclass

from dagzoo.config import (
    SHIFT_MODE_CUSTOM,
    SHIFT_MODE_GRAPH_DRIFT,
    SHIFT_MODE_MECHANISM_DRIFT,
    SHIFT_MODE_MIXED,
    SHIFT_MODE_NOISE_DRIFT,
    SHIFT_MODE_OFF,
    GeneratorConfig,
)
from dagzoo.core.layout_types import MechanismFamily

_LOG_TWO = math.log(2.0)
_NOISE_VARIANCE_DB_SPAN = _LOG_TWO / 2.0

MECHANISM_FAMILY_ORDER: tuple[MechanismFamily, ...] = (
    "nn",
    "tree",
    "discretization",
    "gp",
    "linear",
    "quadratic",
    "em",
    "product",
)
MECHANISM_FAMILY_SUPPORTED_ORDER: tuple[MechanismFamily, ...] = (
    *MECHANISM_FAMILY_ORDER,
    "piecewise",
)

MECHANISM_FAMILY_BASE_LOGITS: dict[MechanismFamily, float] = {
    "nn": 0.7,
    "tree": 0.7,
    "discretization": 0.5,
    "gp": 0.5,
    "linear": -0.8,
    "quadratic": -0.6,
    "em": -0.3,
    "product": 0.9,
    "piecewise": 0.8,
}
NONLINEAR_MECHANISM_FAMILIES: tuple[MechanismFamily, ...] = (
    "nn",
    "tree",
    "discretization",
    "gp",
    "product",
    "piecewise",
)

_MODE_DEFAULT_SCALES: dict[str, tuple[float, float, float]] = {
    SHIFT_MODE_OFF: (0.0, 0.0, 0.0),
    SHIFT_MODE_GRAPH_DRIFT: (0.5, 0.0, 0.0),
    SHIFT_MODE_MECHANISM_DRIFT: (0.0, 0.5, 0.0),
    SHIFT_MODE_NOISE_DRIFT: (0.0, 0.0, 0.5),
    SHIFT_MODE_MIXED: (1.0 / 3.0, 1.0 / 3.0, 1.0 / 3.0),
    SHIFT_MODE_CUSTOM: (0.0, 0.0, 0.0),
}


@dataclass(slots=True, frozen=True)
class ShiftRuntimeParams:
    """Resolved shift runtime parameters for one generation run."""

    enabled: bool
    mode: str
    graph_scale: float
    mechanism_scale: float
    variance_scale: float
    edge_logit_bias_shift: float
    mechanism_logit_tilt: float
    variance_sigma_multiplier: float


def centered_mechanism_family_logits(
    families: tuple[MechanismFamily, ...],
) -> tuple[float, ...]:
    """Return centered family logits used for mechanism drift sampling."""

    if not families:
        return ()
    raw = tuple(float(MECHANISM_FAMILY_BASE_LOGITS.get(name, 0.0)) for name in families)
    mean = sum(raw) / float(len(raw))
    return tuple(value - mean for value in raw)


def mechanism_family_probabilities(
    *,
    mechanism_logit_tilt: float,
    families: tuple[MechanismFamily, ...] | None = None,
    family_weights: dict[MechanismFamily, float] | None = None,
) -> dict[MechanismFamily, float]:
    """Resolve mechanism family probabilities for a given tilt value."""

    if families is None:
        families = (
            MECHANISM_FAMILY_ORDER if family_weights is None else MECHANISM_FAMILY_SUPPORTED_ORDER
        )
    if not families:
        return {}

    if family_weights is None:
        base_weights = {family: 1.0 for family in families}
    else:
        base_weights = {}
        for family in families:
            raw = family_weights.get(family, 0.0)
            if isinstance(raw, bool) or not isinstance(raw, (int, float)):
                raise ValueError(
                    f"mechanism family weight for '{family}' must be a finite number, got {raw!r}."
                )
            parsed = float(raw)
            if not math.isfinite(parsed) or parsed < 0.0:
                raise ValueError(
                    f"mechanism family weight for '{family}' must be finite and >= 0, got {raw!r}."
                )
            base_weights[family] = parsed

    positive_weights = {family: weight for family, weight in base_weights.items() if weight > 0.0}
    if not positive_weights:
        raise ValueError(
            "mechanism family weights must include at least one positive family weight."
        )

    total_base = float(sum(positive_weights.values()))
    normalized_base = {
        family: (positive_weights.get(family, 0.0) / total_base) for family in families
    }
    if mechanism_logit_tilt <= 0.0:
        return normalized_base

    centered_logits = centered_mechanism_family_logits(families)
    scaled = []
    for family, centered in zip(families, centered_logits, strict=True):
        weight = float(normalized_base[family])
        if weight <= 0.0:
            scaled.append(float("-inf"))
        else:
            scaled.append(float(math.log(weight) + (mechanism_logit_tilt * centered)))

    max_logit = max(scaled)
    exp_vals = [math.exp(logit - max_logit) if math.isfinite(logit) else 0.0 for logit in scaled]
    denom = sum(exp_vals)
    if denom <= 0.0:
        raise ValueError("mechanism family probabilities are ill-defined for the provided weights.")
    return {family: (exp_val / denom) for family, exp_val in zip(families, exp_vals, strict=True)}


def mechanism_nonlinear_mass(
    *,
    mechanism_logit_tilt: float,
    families: tuple[MechanismFamily, ...] | None = None,
    nonlinear_families: tuple[MechanismFamily, ...] = NONLINEAR_MECHANISM_FAMILIES,
    family_weights: dict[MechanismFamily, float] | None = None,
) -> float:
    """Return probability mass over nonlinear mechanism families."""

    if families is None:
        families = (
            MECHANISM_FAMILY_ORDER if family_weights is None else MECHANISM_FAMILY_SUPPORTED_ORDER
        )
    if not families:
        return 0.0
    probs = mechanism_family_probabilities(
        mechanism_logit_tilt=mechanism_logit_tilt,
        families=families,
        family_weights=family_weights,
    )
    nonlinear_set = set(nonlinear_families)
    mass = float(math.fsum(prob for family, prob in probs.items() if family in nonlinear_set))
    return min(1.0, max(0.0, mass))


def resolve_shift_runtime_params(config: GeneratorConfig) -> ShiftRuntimeParams:
    """Resolve shift mode/defaults/overrides into runtime coefficients."""

    shift = config.shift
    if not shift.enabled:
        return ShiftRuntimeParams(
            enabled=False,
            mode=SHIFT_MODE_OFF,
            graph_scale=0.0,
            mechanism_scale=0.0,
            variance_scale=0.0,
            edge_logit_bias_shift=0.0,
            mechanism_logit_tilt=0.0,
            variance_sigma_multiplier=1.0,
        )

    mode = str(shift.mode)
    default_graph_scale, default_mechanism_scale, default_noise_scale = _MODE_DEFAULT_SCALES[mode]

    graph_scale = (
        float(shift.graph_scale) if shift.graph_scale is not None else float(default_graph_scale)
    )
    mechanism_scale = (
        float(shift.mechanism_scale)
        if shift.mechanism_scale is not None
        else float(default_mechanism_scale)
    )
    variance_scale = (
        float(shift.variance_scale)
        if shift.variance_scale is not None
        else float(default_noise_scale)
    )

    return ShiftRuntimeParams(
        enabled=True,
        mode=mode,
        graph_scale=graph_scale,
        mechanism_scale=mechanism_scale,
        variance_scale=variance_scale,
        edge_logit_bias_shift=float(_LOG_TWO * graph_scale),
        mechanism_logit_tilt=mechanism_scale,
        variance_sigma_multiplier=float(math.exp(_NOISE_VARIANCE_DB_SPAN * variance_scale)),
    )


__all__ = [
    "MECHANISM_FAMILY_BASE_LOGITS",
    "MECHANISM_FAMILY_ORDER",
    "MECHANISM_FAMILY_SUPPORTED_ORDER",
    "NONLINEAR_MECHANISM_FAMILIES",
    "ShiftRuntimeParams",
    "centered_mechanism_family_logits",
    "mechanism_family_probabilities",
    "mechanism_nonlinear_mass",
    "resolve_shift_runtime_params",
]



"""Categorical converter implementations."""

from __future__ import annotations

from types import SimpleNamespace

import torch

from dagzoo.core.execution_semantics import sample_converter_plan
from dagzoo.core.fixed_layout.batched import FixedLayoutBatchRng, _apply_categorical_group_batch
from dagzoo.core.fixed_layout.plan_types import CategoricalConverterPlan
from dagzoo.core.layout_types import MechanismFamily
from dagzoo.rng import keyed_rng_from_generator


def apply_categorical_converter(
    x: torch.Tensor,
    generator: torch.Generator,
    *,
    n_categories: int,
    method: str | None = None,
    function_family_mix: dict[MechanismFamily, float] | None = None,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Categorical converter in torch via the shared typed-plan semantics."""
    y = x.to(torch.float32)
    if y.dim() == 1:
        y = y.unsqueeze(1)

    c = max(2, int(n_categories))
    root = keyed_rng_from_generator(generator, "apply_categorical_converter")
    plan = sample_converter_plan(
        SimpleNamespace(key="value", kind="cat", dim=int(y.shape[1]), cardinality=c),
        keyed_rng=root.keyed("plan"),
        mechanism_logit_tilt=0.0,
        function_family_mix=function_family_mix,
        method_override=method,
        device=str(generator.device),
    )
    if not isinstance(plan, CategoricalConverterPlan):
        raise RuntimeError("Expected categorical converter plan for categorical converter.")

    rng = FixedLayoutBatchRng.from_keyed_rng(
        root.keyed("execution"),
        batch_size=1,
        device=str(y.device),
    )
    out, labels = _apply_categorical_group_batch(
        y.unsqueeze(0).unsqueeze(2),
        rng,
        plan,
        n_categories=c,
        noise_sigma_multiplier=1.0,
        noise_spec=None,
    )
    return out[0, :, 0, :], labels[0, :, 0]
